/* 1730, Sun 23 Nov 03 (PST)

   FD_DATA.H:  Global declarations for fd_filter and fd_extract

   Copyright (C) 1994-2006 by Russell Fulton & Nevil Brownlee
   University of Auckland | CAIDA*/

/*
 * $Log: fd_data.h,v $
 * Revision 1.1.1.2  2006/03/21 18:08:49  nevil
 * Version 5.1b15
 *
 * Revision 1.1.1.2.2.18  2002/10/31 23:02:25  nevil
 * 45b8: fd_filter 'tail' option implemented, \t supported as format separator
 *
 * Revision 1.1.1.2.2.17  2002/09/13 15:05:11  nevil
 * Implement VLANid, Priority and ECNCodeBits attributes
 *
 * Revision 1.1.1.2.2.16  2002/08/27 17:13:06  nevil
 * Improve thread locking.  Implement STREAM_TO_FLOW, ECNBits and TCPStreamData
 *
 * Revision 1.1.1.2.2.15  2002/07/12 01:41:34  nevil
 * Fix attribute bug in fd_filter
 *
 * Revision 1.1.1.2.2.14  2002/07/03 01:45:45  nevil
 * Implement OAM attributes
 *
 * Revision 1.1.1.2.2.13  2002/06/29 01:55:40  nevil
 * Miscellaneous fixes for warnings found by Compaq Unix cc
 *
 * Revision 1.1.1.2.2.6  2000/08/08 19:44:41  nevil
 * 44b8 release
 *
 * Revision 1.1.1.2.2.4  2000/06/06 03:38:07  nevil
 * Combine NEW_ATR with TCP_ATR, various bug fixes
 *
 * Revision 1.1.1.2.2.1  2000/01/12 02:57:02  nevil
 * Implement 'packet pair matched' turnaroundtime distribution attributes.
 * Fix ASN-related bugs in NeTraMet, distribution-related bugs in fd_filter.
 *
 * Revision 1.1.1.2  1999/10/03 21:06:13  nevil
 * *** empty log message ***
 *
 * Revision 1.1.1.1.2.7  1999/05/18 03:36:23  nevil
 * Implement IPv6 in NeTraMet, and its manager/collectors.
 * - This is controlled by the V6 #define
 * - NeTraMet recognises v6 packets and fishes through their extension
 *     headers until it finds the actual payload.
 * - NeMaC et al display v6 addresses in the fom specified by RFC 2373
 * - fd_util and fd_extract allow colons in addresses (for defining tags)
 *
 * Revision 1.1.1.1.2.6  1999/02/03 04:41:41  nevil
 * Implementation of TCP attributes, part 5
 *
 * Revision 1.1.1.1.2.5  1999/01/27 04:26:11  nevil
 * Minor corrections to fix compiler warnings
 *
 * Revision 1.1.1.1.2.4  1999/01/20 04:01:32  nevil
 * Implementation of TCP attributes, part 4
 *
 * Revision 1.1.1.1.2.3  1999/01/08 01:33:58  nevil
 * Implement TCPdata and DScodepoint attributes
 *
 * Revision 1.1.1.1.2.2  1998/12/23 01:14:10  nevil
 * Distribution file for 4.3b6
 *
 * Revision 1.1.1.1.2.1  1998/11/26 04:13:48  nevil
 * Fix problems in fd_filter and fd_extract when dealing with 'new'
 * attributes, especially those used with NETFLOW. In particular:
 * - Make sure attribs[] is referred to via attr_ix[]
 * - Get the separators correctly from a #Format record (no quotes)
 * - fd_extract wasn't writing out its data for the last sample!
 *
 * Revision 1.1.1.1  1998/11/16 03:57:26  nevil
 * Import of NeTraMet 4.3b3
 *
 * Revision 1.1.1.1  1998/11/16 03:22:00  nevil
 * Import of release 4.3b3
 *
 * Revision 1.1.1.1  1998/10/28 20:31:23  nevil
 * Import of NeTraMet 4.3b1
 *
 * Revision 1.1.3.2  1998/10/18 23:44:03  nevil
 * Added Nicolai's patches, some 'tidying up' of the source
 *
 * Revision 1.1.3.1  1998/10/13 02:48:15  nevil
 * Import of Nicolai's 4.2.2
 *
 * Revision 1.1.1.1  1998/08/24 12:09:28  nguba
 * NetraMet 4.2 Original Distribution
 *
 * Revision 1.4  1998/07/21 00:43:57  rtfm
 * Change attrib numbers for 'New Attribs' I-D
 * First release version of SRL
 *
 * Revision 1.3  1998/05/07 04:28:53  rtfm
 * Implement NetFlowMet, the Cisco NetFlow RTFM meter
 */

#ifndef EXTERN
#define EXTERN  extern
#define DECLARE  0
#define INIT(v)
#else
#define EXTERN
#define DECLARE  1
#define INIT(v)  = v
#endif

#define TESTING    0  /* Trace execution progress */

extern int errno;

EXTERN int snmp_dump_packet INIT(0);

EXTERN int listrules, listformat, quiet, real_out;

#define MXTAGS  200
#define MXCOLS   20

/* fd_filter and fd_extract treat addresses as strings,
   maaximum lengths as follows .. */

#define MAC_ADDR_LEN      6
#define PEER_ADDR_LEN    20
#define TRANS_ADDR_LEN    2

#define RULE_ADDR_LEN    20

#if HAVE_LIBZ  /* struct flow_stats uses gzFile */
# include <zlib.h>
#endif

#if NEW_ATR
#define TCP_DATA_LEN   40
struct tcpstream_data {
   int valid, ToSynRTT, FromSynRTT;
   Bit32 ToDataPoints, FromDataPoints;
   counter64 ToTotalRTT, FromTotalRTT;
   };
struct subflow_data {
   Bit32 n_subflows, mx_active_subflows,
      ToTCPDecrSeq, FromTCPDecrSeq;
   counter64
      ToTCPLenOctets, FromTCPLenOctets,
      ToTCPSeqOctets, FromTCPSeqOctets,
      ToTCPAckOctets, FromTCPAckOctets;
   };

#define DIST_PARAM_LEN  6
#define MXBUCKETS     120
struct distribution {
   unsigned int selector;  /* Attribute number */
/*    mask_params[DIST_PARAM_LEN], value_params[DIST_PARAM_LEN]; */
   counter64 counts[MXBUCKETS+1];

   unsigned int Transform, ScaleFactor; /* 'mask' parameters */
   unsigned int LowerLimit, UpperLimit;
   unsigned int Buckets, Parameter1;  /* 'value' parameters */
   unsigned int Parameter2, Parameter3;

   struct distribution *next;
   };
#define DS_NULL  0  /* Distribution transform type values */
#define DS_LIN   1
#define DS_LOG   2
#define DS_DYN_REQ    4  /* Dynamic: request, linear */
#define DS_DYN_PTS    5  /* Dynamic: values (Buckets = nbr of values) */
#define DS_DYN_BKT    6  /* Dynamic: bucket counts (limits set from data) */
#define DS_DLOG_REQ   7  /* Dynamic: request, log */
#define DS_DLOG_BKT   8  /* Dynamic: bucket counts, log */
#define DS_LIN_RESET  9  /* Linear, reset after read */
#define DS_LOG_RESET 10  /* Log, reset after read */

#define LO_SCALE(s)   (s & 0x0F)  /* LowerLimit scale */
#define HI_SC_DELTA(s)  (s >> 4)  /* UpperLimit scale = lower+delta */

#define tcp_value(a)     (a == FTTCPDATA)
#endif  /* NEW_ATR */

#define Counter64_value(a)   ( (a >= FTUPOCTETS   && a <= FTDOWNPDUS) \
                            || (a >= FTUPOCTETS_D && a <= FTDOWNPDUS_D) )

#define string_value(a)  (attribs[attr_ix[a]].len > 4)
#define STR_LEN   64
#define SESSION_ID_LEN  128

struct flow_info {  /* Accounting flow as viewed by programs */
   unsigned int FlowIndex;
   unsigned int LowInterface, HighInterface;
   unsigned int LowAdjType, HighAdjType;
   unsigned char LowAdjAddress[STR_LEN+1], LowAdjMask[STR_LEN+1],
      HighAdjAddress[STR_LEN+1], HighAdjMask[STR_LEN+1];
   unsigned int
      LowPeerType, HighPeerType;
   unsigned char LowPeerAddress[STR_LEN+1], LowPeerMask[STR_LEN+1],
      HighPeerAddress[STR_LEN+1], HighPeerMask[STR_LEN+1];
   unsigned int
      LowTransType, HighTransType;
   unsigned int   LowTransAddress, LowTransMask,
      HighTransAddress, HighTransMask;
   unsigned int FlowStatus, FlowRuleSet;
   counter64
      FwdPackets,FwdBytes,  BackPackets,BackBytes;
   unsigned int
      FirstTime,LastTime;

   unsigned char UserID[SESSION_ID_LEN+1];
   unsigned int UserCoS, UserInfo;

   unsigned int LowClass, HighClass, FlowClass;
   unsigned int LowKind, HighKind, FlowKind;

   unsigned int DSCodePoint, ECNCodeBits, VLANid, Priority;
   unsigned int PairInterface, ObsInterface;

   unsigned int DNStld, p2pKind;
   unsigned int AppAtr1, AppAtr2, AppAtr3, AppAtr4;

#if NETFLOW
   unsigned int
      LowRouteASN, HighRouteASN,
      LowRoutePrefix, HighRoutePrefix,  MeterId;
#endif
#if NEW_ATR
   struct tcpstream_data tsd;
   struct subflow_data sfd;
   struct distribution *dist;
   unsigned int
      ToLostPDUs, FromLostPDUs,
      ToPQOverflows, FromPQOverflows;
   unsigned int Spare1;
   unsigned int
       UDPStreams, TCPStreams, TwoWayStreams, TotalStreams;
   unsigned int Ext1, Ext2, Ext3, Ext4, Ext5;
   unsigned int minstreamPDUs;
#endif  /* NEW_ATR */


  /* Delta attributes (created by fd_filter) ... */

   counter64
      upbci,   /* To byte rate (bytes since last sample) */
      dnbci,   /* From byte rate */
      uppci,   /* To packet rate */
      dnpci;   /* From packet rate */
   unsigned int
      TagNbr;  /* Tag nbr */
#if NEW_ATR
   struct tcpstream_data tsd_rate;
   struct subflow_data sfd_rate;
   unsigned int
      d_ToLostPDUs, d_FromLostPDUs,
      d_ToPQOverflows, d_FromPQOverflows;
   struct distribution *dist_rate;
   unsigned int 
      d_UDPStreams, d_TCPStreams, d_TwoWayStreams, d_TotalStreams;
#endif  /* NEW_ATR */

   unsigned int
      d_Ext2, d_Ext3, d_Ext4, d_Ext5;
   };

EXTERN struct flow_info curr_flow;


#define AT_DUMMY       0  /* Addr_type values */

#define AT_IP4         1  /* Address Family Numbers */
#define AT_IP6         2
#define AT_CLNS        3
#define AT_OTHER       6  /* 'generic' 802 address family */
#define AT_ETHERNET    7  /* ifType medium */
#define AT_TOKENRING   9  /* ifType medium */
#define AT_NOVELL     11
#define AT_ETHERTALK  12
#define AT_DECNET     13
#define AT_FDDI       15  /* ifType medium */
#define AT_AAL5       49  /* ifType AAL5 over ATM */

#define MX_PROTOCOLS  AT_DECNET

#define AT3_IGNORE     0  /* Addr_type values */
#define AT3_DUMMY      1
#define AT3_ADJACENT   1
#define AT3_IP         2
#define AT3_CLNS       3
#define AT3_IDPR       4
#define AT3_DECNET     5
#define AT3_NOVELL     6
#define AT3_ETHERTALK  7
#define AT3_TRANS      8
#define AT3_IP_FRAG   11
#define AT3_OTHER     12

EXTERN char new_addr_type[]  /* v3 to v4 AddressType translation */
#if DECLARE
   = {
      0,               /*  0 Default */
      0,               /*  1 Adjacent */
      1,               /*  2 IP */
      3,               /*  3 CLNS */
      0,               /*  4 IDPR */
      13,              /*  5 DECnet */
      11,              /*  6 Novell IPX */
      12,              /*  7 EtherTalk */
      0,               /*  8 Transport */
      0,               /*  9 */
      0,               /* 10 */
      1,               /* 11 IP fragments */
      6,               /* 12 Other */
      }
#endif
   ;

#define PT_ICMP       1  /* IP protocol type values */
#define PT_TCP        6
#define PT_UDP       17
#define PT_OSPF      89

#define WNP_FTPDATA  20  /* Well-known tcp/udp port numbers */
#define WNP_FTP      21
#define WNP_TELNET   23
#define WNP_SMTP     25
#define WNP_DOMAIN   53
#define WNP_GOPHER   70
#define WNP_WWW      80
#define WNP_NNTP    119
#define WNP_NTP     123
#define WNP_SNMP    161

#define RF_SET       -2  /* -1 is EOF !!! */
#define RF_RULES     -3
#define RF_ACTIONS   -4
#define RF_FORMAT    -5
#define RF_STATS     -6
#define RF_TAG       -7
#define RF_COLUMN    -8
#define RF_TIME      -9

#define OP_EQ       -41
#define OP_NE       -42

#define EX_SCALE    -51  /* Keywords for fd_extract */
#define EX_KILO     -52
#define EX_MEGA     -53
#define EX_ELAPSED  -54
#define EX_CLOCK    -55
#define EX_SECS     -56
#define EX_MINS     -57
#define EX_HOURS    -58
#define EX_DAYS     -59
#define EX_TSTRING  -60

#define FTFLOWINDEX		1  /* Flow table attribute values */
#define FTFLOWTIMEMARK          2
#define FTFLOWSTATUS		3

#define FTLOWINTERFACE          4
#define FTLOWADJACENTTYPE       5
#define FTLOWADJACENTADDRESS    6
#define FTLOWADJACENTMASK       7
#define FTLOWPEERTYPE		8
#define FTLOWPEERADDRESS	9
#define FTLOWPEERMASK	       10
#define FTLOWTRANSTYPE         11
#define FTLOWTRANSADDRESS      12
#define FTLOWTRANSMASK         13

#define FTHIINTERFACE          14
#define FTHIADJACENTTYPE       15
#define FTHIADJACENTADDRESS    16
#define FTHIADJACENTMASK       17
#define FTHIPEERTYPE	       18
#define FTHIPEERADDRESS	       19
#define FTHIPEERMASK	       20
#define FTHITRANSTYPE          21
#define FTHITRANSADDRESS       22
#define FTHITRANSMASK          23

#define FTPDUSCALE             24
#define FTOCTETSCALE           25
#define FTRULESET              26
#define FTUPOCTETS	       27
#define FTUPPDUS	       28
#define FTDOWNOCTETS	       29
#define FTDOWNPDUS	       30
#define FTFIRSTTIME	       31
#define FTLASTTIME	       32

#define FTLOWSUBSCRIBERID      33
#define FTHISUBSCRIBERID       34
#define FTSESSIONID            35

#define FTSOURCECLASS          36
#define FTDESTCLASS            37
#define FTFLOWCLASS            38
#define FTSOURCEKIND           39
#define FTDESTKIND             40
#define FTFLOWKIND             41

#define FTAPPDNSTLD            98
#define FTAPPP2PKIND           99
#define FTAPPATR4             100
#define FTAPPATR3             101
#define FTAPPATR2             102
#define FTAPPATR1             103
#define N_APP_ATTRS             6

#define FTSPARE1              104
#define FTUDPSTREAMS          105
#define FTTCPSTREAMS          106
#define FTTWOWAYSTREAMS       107
#define FTTOTALSTREAMS        108

#define FTPAIRINTERFACE       109
#define FTOBSINTERFACE        110
#define FTTCPSTREAMDATA       111  /* <TCPStreamData> */

#define FTPRIORITY            117
#define FTDSCODEPOINT         118
#define FTECNCODEBITS         119
#define FTVLANID              120

#define N_NEW_BASIC  12+N_APP_ATTRS  /* New 'basic' attributes */

#define D_O           256  /* Rate attributes (added by fd_filter) */

#define FTUPOCTETS_D	      (D_O+FTUPOCTETS)
#define FTUPPDUS_D	      (D_O+FTUPPDUS)
#define FTDOWNOCTETS_D	      (D_O+FTDOWNOCTETS)
#define FTDOWNPDUS_D	      (D_O+FTDOWNPDUS)
#define FTTAGNBR              (D_O+FTFIRSTTIME)  /* FIRSTTIME, has no _D */

#define FTSPARE1_D            (D_O+FTSPARE1)
#define FTUDPSTREAMS_D        (D_O+FTUDPSTREAMS)
#define FTTCPSTREAMS_D        (D_O+FTTCPSTREAMS)
#define FTTWOWAYSTREAMS_D     (D_O+FTTWOWAYSTREAMS)
#define FTTOTALSTREAMS_D      (D_O+FTTOTALSTREAMS)
#define FTTCPSTREAMDATA_D     (D_O+FTTCPSTREAMDATA)

#define N_RATE_ATRS            11

#if !NEW_ATR
#define N_OLD_ATRS   (FTFLOWKIND + N_NEW_BASIC + N_RATE_ATRS)
#else
#define FTTOLOSTPDUS          121
#define FTFROMLOSTPDUS        122
#define FTTOPQOVERFLOWS       123  /* PQOF */
#define FTFROMPQOVERFLOWS     124

#define FTTOLOSTPDUS_D        (D_O+FTTOLOSTPDUS)
#define FTFROMLOSTPDUS_D      (D_O+FTFROMLOSTPDUS)
#define FTTOPQOVERFLOWS_D     (D_O+FTTOPQOVERFLOWS)  /* PQOF */
#define FTFROMPQOVERFLOWS_D   (D_O+FTFROMPQOVERFLOWS)
#define N_OLD_ATRS   (FTFLOWKIND + N_NEW_BASIC + N_RATE_ATRS + 8)
                                   /* 4+4 OverflowPDUs */
#endif
#define LASTATTRIB   FTFROMPQOVERFLOWS_D  /* Probably redefined below */

#if !NEW_ATR
#define N_DIST_ATRS        0
#else
#define FTDISTRIBUTIONS        65 /* Bit set for each distrib in flow */
#define FTTOPACKETSIZE         66
#define FTFROMPACKETSIZE       67
#define FTTOINTERARRIVALTIME   68
#define FTFROMINTERARRIVALTIME 69
#define FTTOTURNAROUNDTIME     70
#define FTFROMTURNAROUNDTIME   71
#define FTTOBITRATE            72
#define FTFROMBITRATE          73
#define FTTOPDURATE            74
#define FTFROMPDURATE          75

#define FTTOTCPTIME            76
#define FTFROMTCPTIME          77
#define FTTOTCPSIZE            78
#define FTFROMTCPSIZE          79
#define FTTOTCPRATE1           80
#define FTFROMTCPRATE1         81
#define FTTOTCPRATE2           82
#define FTFROMTCPRATE2         83

/* #define FTTOTURNAROUNDTIME4    88
   #define FTFROMTURNAROUNDTIME4  89 */

#define FTTOBYTEVPKTDENSITY    88
#define FTFROMBYTEVPKTDENSITY  89

#define FTTOBYTEDENSITY        84
#define FTFROMBYTEDENSITY      85
#define FTTOPKTDENSITY         86
#define FTFROMPKTDENSITY       87

#define FTTOFLOWOCTETS         90
#define FTFROMFLOWOCTETS       91
#define FTTOFLOWPDUS           92
#define FTFROMFLOWPDUS         93
#define FTFLOWTIME             94

#define FTTOSTREAMS            95
#define FTFROMSTREAMS          96

#define FTSPAREDISTRIB         97

#define LAST_DISTRIB           97  /* 65+32 = 97 */

#define FTTOPACKETSIZE_D          (D_O+FTTOPACKETSIZE)  /* Rates */
#define FTFROMPACKETSIZE_D        (D_O+FTFROMPACKETSIZE)
#define FTTOINTERARRIVALTIME_D    (D_O+FTTOINTERARRIVALTIME)
#define FTFROMINTERARRIVALTIME_D  (D_O+FTFROMINTERARRIVALTIME)
#define FTTOTURNAROUNDTIME_D      (D_O+FTTOTURNAROUNDTIME)
#define FTFROMTURNAROUNDTIME_D    (D_O+FTFROMTURNAROUNDTIME)
#define FTTOBITRATE_D             (D_O+FTTOBITRATE)
#define FTFROMBITRATE_D           (D_O+FTFROMBITRATE)
#define FTTOPDURATE_D             (D_O+FTTOPDURATE)
#define FTFROMPDURATE_D           (D_O+FTFROMPDURATE)

#define FTTOTCPTIME_D             (D_O+FTTOTCPTIME)
#define FTFROMTCPTIME_D           (D_O+FTFROMTCPTIME)
#define FTTOTCPSIZE_D             (D_O+FTTOTCPSIZE)
#define FTFROMTCPSIZE_D           (D_O+FTFROMTCPSIZE)
#define FTTOTCPRATE1_D            (D_O+FTTOTCPRATE1)
#define FTFROMTCPRATE1_D          (D_O+FTFROMTCPRATE1)
#define FTTOTCPRATE2_D            (D_O+FTTOTCPRATE2)
#define FTFROMTCPRATE2_D          (D_O+FTFROMTCPRATE2)

/* #define FTTOTURNAROUNDTIME4_D     (D_O+FTTOTURNAROUNDTIME4)
   #define FTFROMTURNAROUNDTIME4_D   (D_O+FTFROMTURNAROUNDTIME4) */
#define FTTOBYTEVPKTDENSITY_D     (D_O+FTTOBYTEVPKTDENSITY)
#define FTFROMBYTEVPKTDENSITY_D   (D_O+FTFROMBYTEVPKTDENSITY)

#define FTTOBYTEDENSITY_D         (D_O+FTTOBYTEDENSITY)
#define FTFROMBYTEDENSITY_D       (D_O+FTFROMBYTEDENSITY)
#define FTTOPKTDENSITY_D          (D_O+FTTOPKTDENSITY)
#define FTFROMPKTDENSITY_D        (D_O+FTFROMPKTDENSITY)

#define FTTOFLOWOCTETS_D          (D_O+FTTOFLOWOCTETS)
#define FTFROMFLOWOCTETS_D        (D_O+FTFROMFLOWOCTETS)
#define FTTOFLOWPDUS_D            (D_O+FTTOFLOWPDUS)
#define FTFROMFLOWPDUS_D          (D_O+FTFROMFLOWPDUS)
#define FTFLOWTIME_D              (D_O+FTFLOWTIME)

#define FTTOSTREAMS_D             (D_O+FTTOSTREAMS)
#define FTFROMSTREAMS_D           (D_O+FTFROMSTREAMS)

#define FTSPAREDISTRIB_D          (D_O+FTSPAREDISTRIB)

#define LAST_DISTRIB_D            FTSPAREDISTRIB_D

#define DISTRIB_BASE        FTTOPACKETSIZE
#define DISTRIB_ATTRIB(a)   (a >= DISTRIB_BASE && \
                             a <= LAST_DISTRIB)
#define DELTA_DISTRIB_BASE  FTTOPACKETSIZE_D
#define DELTA_DISTRIB(a)    (a >= DELTA_DISTRIB_BASE && \
                             a <= LAST_DISTRIB_D)
#define N_DIST_ATRS         (1 + (LAST_DISTRIB-FTDISTRIBUTIONS)*2)

# if LASTATTRIB < LAST_DISTRIB_D
#  undef LASTATTRIB
#  define LASTATTRIB       LAST_DISTRIB_D  /* max(basic, dist) */
# endif
#endif

#if !NETFLOW
#define N_NF_ATRS        0
#else
#define FTMETERID             112
#define FTLOWROUTEASN         113
#define FTLOWROUTEPREFIX      114
#define FTHIROUTEASN          115
#define FTHIROUTEPREFIX       116

#define NF_ATTRIB(a)  (a >= FTMETERID && a <= FTHIROUTEPREFIX)
#define N_NF_ATRS     (FTHIROUTEPREFIX+1 - FTMETERID)
# if LASTATTRIB < FTHIROUTEPREFIX
#  undef LASTATTRIB
#  define LASTATTRIB    FTHIROUTEPREFIX  /* max(basic, dist, nf) */
# endif
#endif

#define FTTCPDATA             125
#define FTOAMIDENT            126
#define FTOAMDATA             127
#define FTTCPDATA_D           (D_O+FTTCPDATA)
#define N_TCP_ATRS              4
# if LASTATTRIB < FTTCPDATA_D
#  undef LASTATTRIB
#  define LASTATTRIB    FTTCPDATA_D  /* max(basic, dist, nf, tcp) */
# endif

#define EXT_ATR_BASE          128
#define FTHOSTDATA            129
#define FTH_HOSTS             130
#define FTH_STREAMS1          131
#define FTH_STREAMS2          132
#define FTH_STREAMS3          133
#define FTMINSTREAMPDUS       134

/* Make room for another set of distributions ... */
#define FTDISTRIBS2           135 /* Bit set for each distrib2 in flow */
#define FTLTMINSTREAMPDUS     (FTDISTRIBS2+1)
#define FTLTMINSTREAMOCTETS   (FTDISTRIBS2+2)
#define FTTODNSRECSIZE        (FTDISTRIBS2+3)
#define FTFROMDNSRECSIZE      (FTDISTRIBS2+4)

#define LAST_DISTRIB2         (FTDISTRIBS2+4)  /* Max is +32 */
#define LAST_EXT_ATR          LAST_DISTRIB2

#define FTHOSTDATA_D          (D_O+FTHOSTDATA)
#define FTH_HOSTS_D           (D_O+FTH_HOSTS)
#define FTH_STREAMS1_D        (D_O+FTH_STREAMS1)
#define FTH_STREAMS2_D        (D_O+FTH_STREAMS2)
#define FTH_STREAMS3_D        (D_O+FTH_STREAMS3)
#define FTMINSTREAMPDUS_D     (D_O+FTMINSTREAMPDUS)

#define FTLTMINSTREAMPDUS_D   (D_O+FTLTMINSTREAMPDUS)
#define FTLTMINSTREAMOCTETS_D (D_O+FTLTMINSTREAMOCTETS)
#define FTTODNSRECSIZE_D      (D_O+FTTODNSRECSIZE)
#define FTFROMDNSRECSIZE_D    (D_O+FTFROMDNSRECSIZE)

#define LAST_DISTRIB2_D       FTFROMDNSRECSIZE_D
#define LAST_EXT_D_ATR        LAST_DISTRIB2_D

#define DISTRIB2_BASE        FTLTMINSTREAMPDUS
#define DISTRIB2_ATTRIB(a)   (a >= DISTRIB2_BASE && \
                              a <= LAST_DISTRIB2)
#define DELTA_DISTRIB2_BASE  FTLTMINSTREAMPDUS_D
#define DELTA_DISTRIB2(a)    (a >= DELTA_DISTRIB2_BASE && \
                             a <= LAST_DISTRIB2_D)
#define N_DIST2_ATRS         (1 + (LAST_DISTRIB2-FTDISTRIBS2)*2)


# if LASTATTRIB < LAST_EXT_D_ATR
#  undef LASTATTRIB
#  define LASTATTRIB    LAST_EXT_D_ATR  /* max(basic, dist, nf, tcp, ext) */
# endif
#define N_EXT_ATRS  ((LAST_EXT_ATR - EXT_ATR_BASE) *2)

#define N_ATTRIBS  (N_OLD_ATRS + N_DIST_ATRS \
                      + N_NF_ATRS + N_TCP_ATRS + N_EXT_ATRS)
                    /* N_EXT_ATRS includes the extra set of distributions */

typedef union {
   unsigned int *intval;
   unsigned char *charval;
   counter64 *c64val;
#if NEW_ATR
   struct tcpstream_data *tsval;
   struct subflow_data *sfval;
   struct distribution **distval;
#endif
   } val;
struct attrib_info {
   char *name;
   int index;
   int len;
   val value;
   };

EXTERN int attr_ix[1+LASTATTRIB];  /* Index to attribs[] */

#define N_SYNONYMS   9
   /* Synonyms: LastActiveTime,
         ToTurnaroundTime1, FromTurnaroundTime1,
         d_ToTurnaroundTime1, d_FromTurnaroundTime1,
         ToOctetRate, ToPDURate, FromOctetRate, FromPDURate */

#define SZ_ATTRIBS  (1 + N_ATTRIBS + N_SYNONYMS)
                  /* 1: 'null' */ 
EXTERN struct attrib_info attribs[SZ_ATTRIBS]
#if DECLARE
   = {
   "null",                  0,                      0,              NULL,
   "flowindex",             FTFLOWINDEX,            4,              NULL,
   "flowtimemark",          FTFLOWTIMEMARK,         4,              NULL,
   "flowstatus",            FTFLOWSTATUS,           1,              NULL,
   "sourceinterface",       FTLOWINTERFACE,         1,              NULL,
   "sourceadjacenttype",    FTLOWADJACENTTYPE,      1,              NULL,
   "sourceadjacentaddress", FTLOWADJACENTADDRESS,   MAC_ADDR_LEN,   NULL,
   "sourceadjacentmask",    FTLOWADJACENTMASK,      MAC_ADDR_LEN,   NULL,
   "sourcepeertype",        FTLOWPEERTYPE,          1,              NULL,
   "sourcepeeraddress",     FTLOWPEERADDRESS,       PEER_ADDR_LEN,  NULL,
   "sourcepeermask",        FTLOWPEERMASK,          PEER_ADDR_LEN,  NULL,
   "sourcetranstype",       FTLOWTRANSTYPE,         1,              NULL,
   "sourcetransaddress",    FTLOWTRANSADDRESS,      TRANS_ADDR_LEN, NULL,
   "sourcetransmask",       FTLOWTRANSMASK,         TRANS_ADDR_LEN, NULL,

   "destinterface",         FTHIINTERFACE,          1,              NULL,
   "destadjacenttype",      FTHIADJACENTTYPE,       1,              NULL,
   "destadjacentaddress",   FTHIADJACENTADDRESS,    MAC_ADDR_LEN,   NULL,
   "destadjacentmask",      FTHIADJACENTMASK,       MAC_ADDR_LEN,   NULL,
   "destpeertype",          FTHIPEERTYPE,           1,              NULL,
   "destpeeraddress",       FTHIPEERADDRESS,        PEER_ADDR_LEN,  NULL,
   "destpeermask",          FTHIPEERMASK,           PEER_ADDR_LEN,  NULL,
   "desttranstype",         FTHITRANSTYPE,          1,              NULL,
   "desttransaddress",      FTHITRANSADDRESS,       TRANS_ADDR_LEN, NULL,
   "desttransmask",         FTHITRANSMASK,          TRANS_ADDR_LEN, NULL,

   "pduscale",              FTPDUSCALE,             0,              NULL,
   "octetscale",            FTOCTETSCALE,           0,              NULL,
   "flowruleset",           FTRULESET,              1,              NULL,
   "tooctets",              FTUPOCTETS,             8,              NULL,
   "topdus",                FTUPPDUS,               8,              NULL,
   "fromoctets",            FTDOWNOCTETS,           8,              NULL,
   "frompdus",              FTDOWNPDUS,             8,              NULL,
   "firsttime",             FTFIRSTTIME,            4,              NULL,
   "lasttime",              FTLASTTIME,             4,              NULL,
   "lastactivetime",        FTLASTTIME,             4,              NULL,

   "sessionid",             FTLOWSUBSCRIBERID,      PEER_ADDR_LEN,  NULL,
   "usercos",               FTHISUBSCRIBERID,       1,              NULL,
   "userinfo",              FTSESSIONID,            1,              NULL,

   "sourceclass",           FTSOURCECLASS,          1,              NULL,
   "destclass",             FTDESTCLASS,            1,              NULL,
   "flowclass",             FTFLOWCLASS,            1,              NULL,
   "sourcekind",            FTSOURCEKIND,           1,              NULL,
   "destkind",              FTDESTKIND,             1,              NULL,
   "flowkind",              FTFLOWKIND,             1,              NULL,

   "tooctetrate",           FTUPOCTETS_D,           8,              NULL,
   "topdurate",             FTUPPDUS_D,             8,              NULL,
   "fromoctetrate",         FTDOWNOCTETS_D,         8,              NULL,
   "frompdurate",           FTDOWNPDUS_D,           8,              NULL,
   "d_tooctets",            FTUPOCTETS_D,           8,              NULL,
   "d_topdus",              FTUPPDUS_D,             8,              NULL,
   "d_fromoctets",          FTDOWNOCTETS_D,         8,              NULL,
   "d_frompdus",            FTDOWNPDUS_D,           8,              NULL,

   "tagnbr",                FTTAGNBR,               1,              NULL

#if NEW_ATR
   ,
   "tcpstreamdata",         FTTCPSTREAMDATA,        TCP_DATA_LEN,   NULL,
   "d_tcpstreamdata",       FTTCPSTREAMDATA_D,      TCP_DATA_LEN,   NULL
#endif
   ,
   "dnstld",                FTAPPDNSTLD,            TRANS_ADDR_LEN, NULL,
   "p2pkind",               FTAPPP2PKIND,           TRANS_ADDR_LEN, NULL,
   "apppatr4",              FTAPPATR4,              TRANS_ADDR_LEN, NULL,
   "apppatr3",              FTAPPATR3,              TRANS_ADDR_LEN, NULL,
   "apppatr2",              FTAPPATR2,              TRANS_ADDR_LEN, NULL,
   "apppatr1",              FTAPPATR1,              TRANS_ADDR_LEN, NULL,

   "spare1",                FTSPARE1,               TRANS_ADDR_LEN, NULL,
   "udpstreams",            FTUDPSTREAMS,           4,              NULL,
   "tcpstreams",            FTTCPSTREAMS,           4,              NULL,
   "twowaystreams",         FTTWOWAYSTREAMS,        4,              NULL,
   "totalstreams",          FTTOTALSTREAMS,         4,              NULL,
   "d_spare1",              FTSPARE1_D,             TRANS_ADDR_LEN, NULL,
   "d_udpstreams",          FTUDPSTREAMS_D,         4,              NULL,
   "d_tcpstreams",          FTTCPSTREAMS_D,         4,              NULL,
   "d_twowaystreams",       FTTWOWAYSTREAMS_D,      4,              NULL,
   "d_totalstreams",        FTTOTALSTREAMS_D,       4,              NULL,

   "pairinterface",         FTPAIRINTERFACE,        TRANS_ADDR_LEN, NULL,
   "obsinterface",          FTOBSINTERFACE,         1,              NULL,
   "priority",              FTPRIORITY,             1,              NULL,
   "dscodepoint",           FTDSCODEPOINT,          1,              NULL,
   "ecncodebits",           FTECNCODEBITS,          1,              NULL,
   "vlanid",                FTVLANID,               TRANS_ADDR_LEN, NULL

#if defined(NEW_ATR)
   ,
   "tolostpdus",            FTTOLOSTPDUS,           4,              NULL,
   "fromlostpdus",          FTFROMLOSTPDUS,         4,              NULL,
   "d_tolostpdus",          FTTOLOSTPDUS_D,         4,              NULL,
   "d_fromlostpdus",        FTFROMLOSTPDUS_D,       4,              NULL,
   "topqoverflows",         FTTOPQOVERFLOWS,        4,              NULL,
   "frompqoverflows",       FTFROMPQOVERFLOWS,      4,              NULL,
   "d_topqoverflows",       FTTOPQOVERFLOWS_D,      4,              NULL,
   "d_frompqoverflows",     FTFROMPQOVERFLOWS_D,    4,              NULL,

   "distributions",         FTDISTRIBUTIONS,        0,              NULL,
   "topacketsize",          FTTOPACKETSIZE,         DIST_PARAM_LEN, NULL,
   "frompacketsize",        FTFROMPACKETSIZE,       DIST_PARAM_LEN, NULL,
   "tointerarrivaltime",    FTTOINTERARRIVALTIME,   DIST_PARAM_LEN, NULL,
   "frominterarrivaltime",  FTFROMINTERARRIVALTIME, DIST_PARAM_LEN, NULL,
   "toturnaroundtime1",     FTTOTURNAROUNDTIME,     DIST_PARAM_LEN, NULL,
   "fromturnaroundtime1",   FTFROMTURNAROUNDTIME,   DIST_PARAM_LEN, NULL,
   "toturnaroundtime",      FTTOTURNAROUNDTIME,     DIST_PARAM_LEN, NULL,
   "fromturnaroundtime",    FTFROMTURNAROUNDTIME,   DIST_PARAM_LEN, NULL,
   "tobitrate",             FTTOBITRATE,            DIST_PARAM_LEN, NULL,
   "frombitrate",           FTFROMBITRATE,          DIST_PARAM_LEN, NULL,
   "topktrate",             FTTOPDURATE,            DIST_PARAM_LEN, NULL,
   "frompktrate",           FTFROMPDURATE,          DIST_PARAM_LEN, NULL,

   "totcptime",             FTTOTCPTIME,            DIST_PARAM_LEN, NULL,
   "fromtcptime",           FTFROMTCPTIME,          DIST_PARAM_LEN, NULL,
   "totcpsize",             FTTOTCPSIZE,            DIST_PARAM_LEN, NULL,
   "fromtcpsize",           FTFROMTCPSIZE,          DIST_PARAM_LEN, NULL,
   "totcprate1",            FTTOTCPRATE1,           DIST_PARAM_LEN, NULL,
   "fromtcprate1",          FTFROMTCPRATE1,         DIST_PARAM_LEN, NULL,
   "totcprate2",            FTTOTCPRATE2,           DIST_PARAM_LEN, NULL,
   "fromtcprate2",          FTFROMTCPRATE2,         DIST_PARAM_LEN, NULL,

   /*   "toturnaroundtime4",     FTTOTURNAROUNDTIME4,    DIST_PARAM_LEN, NULL,
	"fromturnaroundtime4",   FTFROMTURNAROUNDTIME4,  DIST_PARAM_LEN, NULL, */
   "tobytevpktdensity",     FTTOBYTEVPKTDENSITY,    DIST_PARAM_LEN, NULL,
   "frombytevpktdensity",   FTFROMBYTEVPKTDENSITY,  DIST_PARAM_LEN, NULL,

   "tobytedensity",         FTTOBYTEDENSITY,        DIST_PARAM_LEN, NULL,
   "frombytedensity",       FTFROMBYTEDENSITY,      DIST_PARAM_LEN, NULL,
   "topktdensity",          FTTOPKTDENSITY,         DIST_PARAM_LEN, NULL,
   "frompktdensity",        FTFROMPKTDENSITY,       DIST_PARAM_LEN, NULL,
   "tostreams",             FTTOSTREAMS,            DIST_PARAM_LEN, NULL,
   "fromstreams",           FTFROMSTREAMS,          DIST_PARAM_LEN, NULL,

   "toflowoctets",          FTTOFLOWOCTETS,         DIST_PARAM_LEN, NULL,
   "fromflowoctets",        FTFROMFLOWOCTETS,       DIST_PARAM_LEN, NULL,
   "toflowpdus",            FTTOFLOWPDUS,           DIST_PARAM_LEN, NULL,
   "fromflowpdus",          FTFROMFLOWPDUS,         DIST_PARAM_LEN, NULL,
   "flowtime",              FTFLOWTIME,             DIST_PARAM_LEN, NULL,

   "sparedistrib",          FTSPAREDISTRIB,         DIST_PARAM_LEN, NULL,

   "d_topacketsize",         FTTOPACKETSIZE_D,         DIST_PARAM_LEN, NULL,
   "d_frompacketsize",       FTFROMPACKETSIZE_D,       DIST_PARAM_LEN, NULL,
   "d_tointerarrivaltime",   FTTOINTERARRIVALTIME_D,   DIST_PARAM_LEN, NULL,
   "d_frominterarrivaltime", FTFROMINTERARRIVALTIME_D, DIST_PARAM_LEN, NULL,
   "d_toturnaroundtime1",    FTTOTURNAROUNDTIME_D,     DIST_PARAM_LEN, NULL,
   "d_fromturnaroundtime1",  FTFROMTURNAROUNDTIME_D,   DIST_PARAM_LEN, NULL,
   "d_toturnaroundtime",     FTTOTURNAROUNDTIME_D,     DIST_PARAM_LEN, NULL,
   "d_fromturnaroundtime",   FTFROMTURNAROUNDTIME_D,   DIST_PARAM_LEN, NULL,
   "d_tobitrate",            FTTOBITRATE_D,            DIST_PARAM_LEN, NULL,
   "d_frombitrate",          FTFROMBITRATE_D,          DIST_PARAM_LEN, NULL,
   "d_topktrate",            FTTOPDURATE_D,            DIST_PARAM_LEN, NULL,
   "d_frompktrate",          FTFROMPDURATE_D,          DIST_PARAM_LEN, NULL,

   "d_totcptime",            FTTOTCPTIME_D,            DIST_PARAM_LEN, NULL,
   "d_fromtcptime",          FTFROMTCPTIME_D,          DIST_PARAM_LEN, NULL,
   "d_totcpsize" ,           FTTOTCPSIZE_D,            DIST_PARAM_LEN, NULL,
   "d_fromtcpsize",          FTFROMTCPSIZE_D,          DIST_PARAM_LEN, NULL,
   "d_totcprate1",           FTTOTCPRATE1_D,           DIST_PARAM_LEN, NULL,
   "d_fromtcprate1",         FTFROMTCPRATE1_D,         DIST_PARAM_LEN, NULL,
   "d_totcprate2",           FTTOTCPRATE2_D,           DIST_PARAM_LEN, NULL,
   "d_fromtcprate2",         FTFROMTCPRATE2_D,         DIST_PARAM_LEN, NULL,

   /*   "d_toturnaroundtime4",   FTTOTURNAROUNDTIME4_D,     DIST_PARAM_LEN, NULL,
	"d_fromturnaroundtime4", FTFROMTURNAROUNDTIME4_D,   DIST_PARAM_LEN, NULL, */
   "d_tobytevpktdensity",    FTTOBYTEVPKTDENSITY_D,    DIST_PARAM_LEN, NULL,
   "d_frombytevpktdensity",  FTFROMBYTEVPKTDENSITY_D,  DIST_PARAM_LEN, NULL,

   "d_tobytedensity",        FTTOBYTEDENSITY_D,        DIST_PARAM_LEN, NULL,
   "d_frombytedensity",      FTFROMBYTEDENSITY_D,      DIST_PARAM_LEN, NULL,
   "d_topktdensity",         FTTOPKTDENSITY_D,         DIST_PARAM_LEN, NULL,
   "d_frompktdensity",       FTFROMPKTDENSITY_D,       DIST_PARAM_LEN, NULL,
   "d_tostreams",            FTTOSTREAMS_D,            DIST_PARAM_LEN, NULL,
   "d_fromstreams",          FTFROMSTREAMS_D,          DIST_PARAM_LEN, NULL,

   "d_toflowoctets",         FTTOFLOWOCTETS_D,         DIST_PARAM_LEN, NULL,
   "d_fromflowoctets",       FTFROMFLOWOCTETS_D,       DIST_PARAM_LEN, NULL,
   "d_toflowpdus",           FTTOFLOWPDUS_D,           DIST_PARAM_LEN, NULL,
   "d_fromflowpdus",         FTFROMFLOWPDUS_D,         DIST_PARAM_LEN, NULL,
   "d_flowtime",             FTFLOWTIME_D,             DIST_PARAM_LEN, NULL,

   "d_sparedistrib",         FTSPAREDISTRIB_D,         DIST_PARAM_LEN, NULL
#endif

#if NETFLOW
   ,
   "meterid",                FTMETERID,              1,              NULL,
   "sourceasn",              FTLOWROUTEASN,          TRANS_ADDR_LEN, NULL,
   "sourceprefix",           FTLOWROUTEPREFIX,       1,              NULL,
   "destasn",                FTHIROUTEASN,           TRANS_ADDR_LEN, NULL,
   "destprefix",             FTHIROUTEPREFIX,        1,              NULL
#endif

#if NEW_ATR
   ,
   "tcpdata",                FTTCPDATA,              TCP_DATA_LEN,   NULL,
   "d_tcpdata",              FTTCPDATA_D,            TCP_DATA_LEN,   NULL,
   "oamident",               FTOAMIDENT,             1,              NULL,
   "oamdata",                FTOAMDATA,              DIST_PARAM_LEN, NULL
#endif

   ,
   "hostdata",               FTHOSTDATA,             TRANS_ADDR_LEN, NULL,
   "h_hosts",                FTH_HOSTS,              TRANS_ADDR_LEN, NULL,
   "h_streams1",             FTH_STREAMS1,           TRANS_ADDR_LEN, NULL,
   "h_streams2",             FTH_STREAMS2,           TRANS_ADDR_LEN, NULL,
   "h_streams3",             FTH_STREAMS3,           TRANS_ADDR_LEN, NULL,
   "minstreampdus",          FTMINSTREAMPDUS,        1,              NULL,

   "d_hostdata",             FTHOSTDATA_D,           TRANS_ADDR_LEN, NULL,
   "d_h_hosts",              FTH_HOSTS_D,            TRANS_ADDR_LEN, NULL,
   "d_h_streams1",           FTH_STREAMS1_D,         TRANS_ADDR_LEN, NULL,
   "d_h_streams2",           FTH_STREAMS2_D,         TRANS_ADDR_LEN, NULL,
   "d_h_streams3",           FTH_STREAMS3_D,         TRANS_ADDR_LEN, NULL,
   "d_minstreampdus",        FTMINSTREAMPDUS_D,      1,              NULL,

   "distributions2",         FTDISTRIBS2,            0,              NULL,
   "ltminstreampdus",        FTLTMINSTREAMPDUS,      DIST_PARAM_LEN, NULL,
   "ltminstreamoctets",      FTLTMINSTREAMOCTETS,    DIST_PARAM_LEN, NULL,
   "todnsrecordsize",        FTTODNSRECSIZE,         DIST_PARAM_LEN, NULL,
   "fromdnsrecordsize",      FTFROMDNSRECSIZE,       DIST_PARAM_LEN, NULL,

   "d_ltminstreampdus",      FTLTMINSTREAMPDUS_D,    DIST_PARAM_LEN, NULL,
   "d_ltminstreamoctets",    FTLTMINSTREAMOCTETS_D,  DIST_PARAM_LEN, NULL,
   "d_todnsrecordsize",      FTTODNSRECSIZE_D,       DIST_PARAM_LEN, NULL,
   "d_fromdnsrecordsize",    FTFROMDNSRECSIZE_D,     DIST_PARAM_LEN, NULL
   }
#endif
   ;

struct flow_stats {
   int format[1+N_ATTRIBS];
   char *separator[1+N_ATTRIBS];
   int n_attr;
   int required[1+LASTATTRIB];
#if NEW_ATR
  unsigned long distrib_bits, delta_bits,
      distrib2_bits, delta2_bits;
#endif  /* NEW_ATR */

   FILE *flows;
   int MIBver;

   int zfile;
#if HAVE_LIBZ
   gzFile gz_flows;
#endif
   char *filename;
   };


EXTERN FILE *rfp;  /* Scanner globals */
EXTERN char inbuf[12000],*ibp, ebuf[80];  /* Input, error message buffers */
EXTERN int lic, ic,  /* Last input char, current input char */
   iblisted, rule_line, rferrors;

EXTERN char outbuf[12000],*obp;  /* Outpuf buffer */

/* Forward procedure declarations .. */

/* Functions declared in fd_parse.c */

char *gnbr(unsigned int *n, char *s);
char *gcstring(char *s, int *len);

int parse_open(char *fn);
void p_error(char *msg);
int nextchar(void);
int nexttoken(void);
int relop_next(void);
void getfdstring(unsigned char *arg);
int wordis(char *p, char *w);
int getword(int res_reqd);
int getnbr(void);

int open_flow_file(struct flow_stats *fs, char *filename, char *mode);
void close_flow_file(struct flow_stats *fs);
int get_flow_record(struct flow_stats *fs);
void put_flow_record(struct flow_stats *fs);

void scan_format(struct flow_stats *fs, int single_record, int list);

val attrib_ptr(struct flow_info *fip, int a);
char *get_value(val v, int a, char *pb);
void put_value(int a, val v);

#if NEW_ATR
struct distribution *get_dist(void);
void free_dist(struct distribution *d);
#endif  /* NEW_ATR */

char *strmov(char *d, char *s);
char *fmt_time(time_t *ti, int utc);
char *str_time(char *fts);

/* Globals for fd_support.c  . . */

#define DATA_REC  1
#define TIME_REC  2
#define STAT_REC  3
#define END_DATA  4
#define UNK_REC   5

struct tag_info {
   unsigned char  /* Attributes to test, null terminated */
      att_nbr[1+LASTATTRIB];
   int cmp_eq[1+LASTATTRIB];
   val att_p[1+LASTATTRIB];  /* Pointers to attribute values */

   struct flow_info flow;  /* Attribute values */
   };

#if NEW_ATR
EXTERN int tcpdata_in_format INIT(0);
#endif

struct flow_data {  /* Holds values from earlier sample */
   unsigned short ruleset;
   unsigned int flownum, starttime;
   counter64
      upbc, dnbc, uppc, dnpc;
#if NEW_ATR
   struct tcpstream_data tsd;
   struct subflow_data sfd;
   unsigned int
      ToLostPDUs, FromLostPDUs,
      ToPQOverflows, FromPQOverflows,
      StreamCount;
   struct distribution *dist_rate;
   unsigned int
       UDPStreams, TCPStreams, TwoWayStreams, TotalStreams;
#endif

   unsigned int Ext2, Ext3, Ext4, Ext5;
   unsigned int minstreamPDUs;
   };
EXTERN struct flow_data *flow;


void quack(int n);  /* For gdb */
struct tag_info *scan_tag(void);
void list_tags(void);
void distrib_deltas(struct distribution *d_rate,
   struct distribution *cf_dist, struct distribution *fi_dist);
void write_filter_info(int argc, char *argv[], int utc,
   struct flow_stats *out_set);
void parse_format(char *format_fn, struct flow_stats *flow_set);
int process_header(int *utc,
   struct flow_stats *in_set, struct flow_stats *out_set);
int process_flow(char *inbuf, struct flow_stats *in_set,
   struct flow_stats *out_set);
void read_format_rec(
   struct flow_stats *in_set, struct flow_stats *out_set);
static void write_flow_rec(struct flow_stats *out_set);
void alloc_flows(struct flow_stats *in_set,
   struct flow_stats *out_set, int max_flows);
int process_rec(char *inbuf, struct flow_stats *in_set,
   struct flow_stats *out_set);
void init_dists(struct flow_stats *flow_set);
void init_attribs(void);
void  free_flow_mem (int max_flows);
int check_formats(struct flow_stats *in, struct flow_stats *out);
