/* 1145, Wed 10 Mar 04 (NZDT)

   nsnap.c:  Nevil's simple version of dagsplit

   Copyright (C) 1992-2006 by Nevil Brownlee,
   CAIDA | University of Auckland */

/*
 * $Log: nsnap.c,v $
 * Revision 1.1.1.1  2006/03/21 18:08:53  nevil
 * Version 5.1b15
 *
 * Revision 1.1.1.2.2.19  2002/10/31 23:02:37  nevil
 * 45b8: fd_filter 'tail' option implemented, \t supported as format separator
 * 
 */

#if HAVE_CONFIG_H
#include <ntm_conf.h>
#endif

#define DAG_EXT_TESTING  0  /* For debugging varlen slen=nn */
#define DEBUG_TF_READ    0  /* For debugging trace-file input */

#include <sys/types.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#if HAVE_SYS_SELECT_H
#include <sys/select.h>
#endif
#include <sys/utsname.h>
#include <arpa/inet.h>

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <stdarg.h>
#include <signal.h>

#include <time.h>
#include <sys/time.h>
#include <sys/times.h>
#include <limits.h>
//#include <sys/sysinfo.h>

#ifdef SUNOS
#include <stropts.h>
#define void
#define const
#endif

#if MULTI
# include <pthread.h>
# include "threads/semaphore.h"
#endif

#define EXTSNMP
#include "ausnmp.h"

#define PKTSNAP
#include "pktsnap.h"

#define MXINTERFACES  4  /* Max nbr of interfaces to meter */

#if HAVE_LIBZ
# include <zlib.h>
#endif

#if DAG_DIRECT
# include <dagnew.h>

# if DAG_NEW_API
#  include <sys/mman.h>
#  include <dagapi.h>
# endif

/* Dag trace file record layouts ...
      Note that the 'legacy' formats use 64-byte fixed-length records */

struct dag_atm {  /* Dag record for ATM (first cell only, a la OC3MON) */
   Bit32 crc;           /* Always zero */
   Bit32 atm_hdr;       /* VPI:VCI, etc */
   struct snap_hdr sh;  /* LLC snap header (8 bytes) */
   Bit8 pload[40];      /* First 40 bytes of IP packet */
   };

struct dag_ether {  /* Dag record for Ethernet */
   Bit16 wlen;          /* Wire length (bytes) */
   struct ether_hdr eh; /* Ethernet MAC header */
   Bit8 pload[40];
   };

struct dag_pos {  /* Dag record for PoS */
   Bit32 rlen;          /* Snap length, must be 64 for now */
   Bit16 llctr;         /* Loss counter (due toPCI bus overload) */
   Bit16 wlen;          /* Bytes on the wire, seems to include FCS */
   Bit8 chdlc;          /* Cisco HDLC header (this + next 3 bytes) */
   Bit8 unused;         /* Padding */
   Bit16 etype;         /* EtherType */
   Bit8 pload[44];      /* 4 more bytes than ATM and Ether */
   };

struct dag_ext_pos {  /* Extensible Type 1, PoS HDLC */
   Bit8 chdlc;          /* Cisco HDLC header (this + next 3 bytes) */
   Bit8 unused;         /* Padding */
   Bit16 etype;         /* EtherType */
   Bit8 pload[44];      /* rlen-20, rlen default is 64 */
   };

struct dag_ext_ether {  /* Extensible Type 2, Ethernet */
   Bit8 offset;         /* Bytes to skip (not captured) */
   Bit8 pad;            /* Reserved */
   struct ether_hdr eh; /* Ethernet MAC header */
   Bit8 pload[32];      /* rlen-18 */
   };

struct dag_ext_atm {  /* Dag record for ATM (first cell only, a la OC3MON) */
   Bit32 atm_hdr;       /* VPI:VCI, etc */
   struct snap_hdr sh;  /* LLC snap header (8 bytes) */
   Bit8 pload[40];      /* Data bytes for cell */
   };

struct dag_ext {      /* 'Extensible' Dag record */
   Bit8 type;           /* Dag type (TYPE_ defines, in dag.h) */
   Bit8 flags;          /* Dag flags (DF_ defines, below) */
   Bit16 rlen;          /* Dag data length */
   Bit16 lctr;          /* Loss counter, sticks at 0xffff */
   Bit16 wlen;          /* Bytes on the wire */
   union {
      struct dag_ext_pos pos;
      struct dag_ext_ether eth;
      struct dag_ext_atm atm;
      } e;
   };

#define DF_INTERFACE  0x03  /* Capture interface 0-3 */
#define DF_VBL_LEN    0x04  /* Variable length records */
#define DF_TRUNCATED  0x08  /* Record truncated (insufficient buffer space) */
#define DF_RX_ERROR   0x10  /* Rx error (link layer) */
#define DF_DS_ERROR   0x20  /* Internal error */
#define DF_RSRV       0xC0  /* Reserved */

#define TT_ATM    0  /* Dag trace file record types */
#define TT_ETHER  1
#define TT_POS    2
#define TT_EXT    3

#define TYPE_LEGACY    0  /* Values for dag_ext type field */
#define TYPE_HDLC_POS  1
#define TYPE_ETH       2
#define TYPE_ATM       3

#define DAG_DEFAULT_SLEN   48  /* These match the Legacy records */
#define DAG_DEFAULT_VARLEN  0

struct dag_rec {
   long long ts;        /* Dag timestamp (64-bit) */
   union {
      struct dag_atm atm;
      struct dag_ether eth;
      struct dag_pos pos;
      struct dag_ext ext;
      } d;
   };

# if MULTI
#  define N_DAG_BUFS  8*16384/DAG_BUF_SZ  /* * MB per interface */
# else
#  define N_DAG_BUFS  1  /* Only one for single-thread meter */
# endif

#endif  /* DAG_DIRECT */


#define SZ_IFNAME  64
struct interface_info {
   u_char nbr;  /* Interface number (for Surce/DestInterface attributes) */
   u_char adjtype;  /* Adjacent type for interface */
   char name[SZ_IFNAME+1];  /* Interface name, e.g. nf0 */
   u_int SampleRate,  /* 1 = count every packet on this interface */
      sample_count;
   int fd, last_errno;
   u_int ipackets;  /* Total packets */
   u_int idrops;  /* Dropped packets */
   u_int noflowpackets;

   int nblocks,last_nblocks;
   int nrecords;  /* For osp() 5-min log records */
   u_int badtspkts;  /* Packets with bad timestamps */
   u_int no_data_warned;
   int last_s;  /* State from last read */
#     define S_NORMAL  0
#     define S_EMPTY   1
#     define S_FULL    2
   u_int full,empty, mx_full,mx_empty;
   int consec_badts, restarts;

# if DAG_DIRECT
#  if HAVE_LIBZ
    gzFile gfp;  /* For reading Dag trace files */
#else 
    int t_fd;
#  endif
#   define DF_BLK_SZ  0x10000
// #   define DF_BLK_SZ  1024
   Bit8 *fbuf;
   int
      fbr,  /* Nbr of fbuf bytes read in last block */
      fbx;  /* Index in fbuf of first byte of current record */

   int dag_type, dag_slen;
   int dag_ports;
#  if DAG_NEW_API
   int dag_varlen;
   void *dbuf;  /* Dag driver buffer */
   unsigned bottom;  /* Offset of next byte to read */
   unsigned n_stuck_lctrs;  /* Nbr of time lctr reached 0xFFFF */
   struct dag_rec *dbp;
   int r_pkts;
#  elif
   Bit8 *in_buf[N_DAG_BUFS];  /* Array of buf pointers */
   int in_bytes[N_DAG_BUFS];
   int inx, outx;
#  endif
   int running;
   int r_bytes;  /* Bytes left in this block */
   int n_pkts;  /* Packets found so far in this block */
   struct dag_rec *drp;  /* Current record in this block */
   long long last_read_ts;  /* First ts of last block */
   long long card_ts;  /* Last ts of last block */

# endif
   };

struct interface_info *pci[MXINTERFACES];

int live_source = 1;  /* Set to zero to use trace files */

static int crl_n_max = 0;  /* Set by -N option */
static int crl_n_count;    /* Counting to crl_n_max */
static int trace_count;    /* Counter for 1s intervals within trace_interval */

static unsigned long
    npackets_org = 0L, lostpackets_org = 0L, 
    pkts = 0L, drops = 0L;

#if DAG_DIRECT
long long meter_start_time = 0;  /* Dag (64-bit) timestamp */
#else
struct timeval meter_start_time = {0, 0};
#endif

/* Make SIGINT or SIGTERM shut the meter down gracefully */

int request_stop = 0;
void sigint_handler(int x)  /* INT (keyboard) and TERM (kill) */
{                /* To make ^c send INT, use   stty intr '^c' */
   request_stop = 1;
   }

int got_USR1 = 0;
void sigusr1_handler(int x)  /* kill -USR1 */
{
   got_USR1 = !got_USR1;
   }

int got_USR2 = 0;
void sigusr2_handler(int x)  /* kill -USR2 */
{
   got_USR2 = !got_USR2;
   }

#if MULTI
#if 0
#define PME_BUF_RECS     59
#define PME_BUF_RECS    113
#define PME_BUF_RECS    493
#define PME_BUF_RECS   1019
#define PME_BUF_RECS   2029
#define PME_BUF_RECS   4091
#define PME_BUF_RECS   8187
#define PME_BUF_RECS  16369
#endif

#if MULTI_BUF_SZ == 3
#define PME_BUF_RECS    113
#elif MULTI_BUF_SZ == 2
#define PME_BUF_RECS     59
#elif MULTI_BUF_SZ == 1
#define PME_BUF_RECS     23
#else  /* For testing on low-bit-rate links */
#define PME_BUF_RECS      2
#warning ---> PME_BUF_RECS  2 <----
#endif

#if LOG_THREAD_PROCTIMES
FILE *threads_log;
#define MAX_THREADS  12
#define THR_NAME_SZ  20
int assign_thr_log_nbr(void);
char thread_name[MAX_THREADS][THR_NAME_SZ];
Bit32 thread_proctime[MAX_THREADS];
Bit32 thread_proctime_org[MAX_THREADS];

int t_log_n = 0;
int assign_thr_log_nbr(void)
{ return t_log_n++; }
#endif

# if THREAD_GPROF
/* Following wrapper code to make gprof work properly for threads
   is from   http://sam.zoy.org/doc/programming/gprof.html  */

/* 
 * pthread_create wrapper for gprof compatibility
 *
 * needed headers: <pthread.h>
 *                 <sys/time.h>
 */

typedef struct wrapper_s {
    void * (*start_routine)(void *);
    void * arg;

    pthread_mutex_t lock;
    pthread_cond_t  wait;

    struct itimerval itimer;
    } wrapper_t;

static void * wrapper_routine(void *);

/* Same prototype as pthread_create; use some #define magic to
 * transparently replace it in other files */
int gprof_pthread_create(pthread_t * thread, pthread_attr_t * attr,
			 void * (*start_routine)(void *), void * arg)
{
   wrapper_t wrapper_data;
   int i_return;

   /* Initialize the wrapper structure */
   wrapper_data.start_routine = start_routine;
   wrapper_data.arg = arg;
   getitimer(ITIMER_PROF, &wrapper_data.itimer);
   pthread_cond_init(&wrapper_data.wait, NULL);
   pthread_mutex_init(&wrapper_data.lock, NULL);
   pthread_mutex_lock(&wrapper_data.lock);

   /* The real pthread_create call */
   i_return = pthread_create(thread, attr, &wrapper_routine,
			    &wrapper_data);

   /* If the thread was successfully spawned, wait for the data
    * to be released */
   if(i_return == 0) {
      pthread_cond_wait(&wrapper_data.wait, &wrapper_data.lock);
      }

   pthread_mutex_unlock(&wrapper_data.lock);
   pthread_mutex_destroy(&wrapper_data.lock);
   pthread_cond_destroy(&wrapper_data.wait);

   return i_return;
   }

/* The wrapper function in charge for setting the itimer value */
static void * wrapper_routine(void * data)
{
   /* Put user data in thread-local variables */
   void * (*start_routine)(void *) = ((wrapper_t*)data)->start_routine;
   void * arg = ((wrapper_t*)data)->arg;

   /* Set the profile timer value */
   setitimer(ITIMER_PROF, &((wrapper_t*)data)->itimer, NULL);

   /* Tell the calling thread that we don't need its data anymore */
   pthread_mutex_lock(&((wrapper_t*)data)->lock);
   pthread_cond_signal(&((wrapper_t*)data)->wait);
   pthread_mutex_unlock(&((wrapper_t*)data)->lock);

   /* Call the real function */
   return start_routine(arg);
   }
# else
# define gprof_pthread_create(thread, ta, f, arg)  \
      pthread_create(thread, ta, f, arg)
# endif
#endif


#if LINUX
# include <unistd.h>
# include <sys/sysinfo.h>
#elif SOLARIS
#include <unistd.h>
#elif FREEBSD || NETBSD
# include <sys/types.h>
# include <sys/sysctl.h>
#elif ALPHA
# include <sys/sysinfo.h>
#endif

int n_procs;  /* Nbr of processors configured on system */
unsigned phys_mem;  /* Physical memory (bytes) */

void get_sys_info(int hyperthread_processors)
{
#if LINUX
   phys_mem = sysconf(_SC_PAGE_SIZE) * sysconf(_SC_PHYS_PAGES);
   n_procs = get_nprocs_conf();
#elif SOLARIS
   phys_mem = sysconf(_SC_PAGE_SIZE) * sysconf(_SC_PHYS_PAGES);
   n_procs = sysconf(_SC_NPROCESSORS_ONLN);
#elif FREEBSD || NETBSD
  int mib[2];
  size_t len = sizeof(unsigned int);

  mib[0] = CTL_HW;
  mib[1] = HW_NCPU;
  sysctl(mib, 2, &n_procs, &len, NULL, 0);
  mib[1] = HW_PHYSMEM;
  sysctl(mib, 2, &phys_mem, &len, NULL, 0);
#elif ALPHA
   long pm;
   getsysinfo(GSI_PHYSMEM, (char *)&pm, sizeof(pm), NULL,NULL,NULL);
   phys_mem = (int)pm * 1024; 
   getsysinfo(GSI_CPUS_IN_BOX, (char *)&n_procs, sizeof(n_procs), NULL,NULL,NULL);
#endif
   if (hyperthread_processors) n_procs /= 2;
      /* Linux counts Intel Xeon chips as two processors - 
         although they run two threads each, they only have
         one actual processor! */
   } 


void stop_meter(int code)
{
   display_msg(1,"Shutting down");
#if CR_DATA
   coral_stop_all();  /* Leave Dag cards in a graceful state */
   coral_close_all();
#endif
#if DAG_NEW_API
   if (live_source)
      dag_close_all();
#endif
#if MULTI
   destroy_mutexes();
#endif
   exit(code);
   }



#if DAG_DIRECT  /* !NF_CISCO_DATA && !CR_DATA && !LFAP_METER && !ADSL_DATA */

/* ===== Dag direct interface code ===== */

#include <fcntl.h>

#define MERGE_TEST  0

#define MX_CONSEC_BADTS  25
   /* Shut down meter after this many bad timestamps */

#define MX_TS_JUMP      2  /* Seconds */

int waiting_live_source = -1;

int select_us = 4*MN_SELECT_US;
long long select_time;
#define MX_NODATA_READS  100

void fmt_ts(char *t_buf, ts_type t)
{
   long long mt = t - meter_start_time;
   char *neg = "";
   if (mt < 0) {
      mt = -mt;  neg = "-";
      }
   sprintf(t_buf, "%s%lu.%06lu", neg, DAGT_S(mt),DAGT_US(mt));
   }

#if PKT_DEBUG > 1
int n_rdr_pr = 0;
#endif

struct pkt *pkt_from_dag_record(struct dag_rec *drp,
   struct interface_info *pi)
{
   Bit32 ah, vpvcmask;
   struct snap_hdr *sh;
   struct ether_hdr ehdr, *eh;
   Bit8 *p;  int rec_len;
   Bit16 lsap, ether_type;
   int t, r;
   struct pkt *p_p;

   if ((p_p = get_pkt()) == NULL)
      return NULL;  /* No free pkts, drop the record */

# if PKT_DEBUG
   if (p_p->pkt_state != 1) {
      printf("pkt_from_dag(1): not got, p_p=%x\n", p_p);  fflush(stdout);
      quack(401);
      }
# endif
# if PKT_DEBUG > 1
   printf("rdr(%x) ", p_p);
   if (++n_rdr_pr == 8) { printf("\n");  n_rdr_pr = 0; }
   fflush(stdout);
# endif

# if 0
   { char tsb[50];
   fmt_ts(tsb, drp->ts);  /**/
     log_msg(LOG_ERR, 0, "thd(%d): ts=%s, ncells=%d",
	     pi->nbr, tsb, pi->ncells); } /**/
# endif

   p_p->arrival_time = microseconds(drp->ts);
   if (adj_reqd) memset(&ehdr, 0, sizeof(struct ether_hdr));
   p_p->if_nbr = pi->nbr;
   p_p->n_ports = pi->dag_ports;

# if PKT_DEBUG
   if (p_p->pkt_state != 1) {
      printf("pkt_from_dag(2): not got, p_p=%x\n", p_p);  fflush(stdout);
      quack(402);
      }
# endif
   if ((t = drp->d.ext.type) == TYPE_HDLC_POS) {
      p_p->port_nbr = drp->d.ext.flags & DF_INTERFACE;
      p_p->p_len = ntohl(drp->d.ext.wlen);  /* On-the-wire length */
      p_p->adj_type = AT_SONET;
      rec_len = ntohs(drp->d.ext.rlen) -  /* Bytes in Dag record */
	(drp->d.ext.e.pos.pload - (Bit8 *)drp);
      r = handle_pkt(p_p, ntohs(drp->d.ext.e.pos.etype), 0, (Bit8 *)&ehdr,
         drp->d.ext.e.pos.pload, rec_len);
      }
   else if (t == TYPE_ETH) {
      p_p->port_nbr = drp->d.ext.flags & DF_INTERFACE;
      p_p->p_len = ntohl(drp->d.ext.wlen);  /* On-the-wire length */
      p_p->adj_type = AT_ETHERNET;
      rec_len = ntohs(drp->d.ext.rlen) -  /* Bytes in Dag record */
         (drp->d.ext.e.eth.pload - (Bit8 *)drp);
      p = drp->d.ext.e.eth.pload;
      eh = &drp->d.ext.e.eth.eh;
      if ((ether_type = ntohs(eh->type)) <= 1500) {  /* 802.3 packet */
         sh = (struct snap_hdr *)p;
         if ((lsap = ntohs(sh->lsap)) == 0xAAAA) {  /* SNAP packet */
            ether_type = ntohs(sh->etype);
            p += sizeof(struct snap_hdr);
            rec_len -= sizeof(struct snap_hdr);
            }
         else ether_type = 0;  /* Other 802.2 packet */
         }
      else lsap = 0;  /* Blue book packet */
      r = handle_pkt(p_p, ether_type, lsap, (Bit8 *)&ehdr, p, rec_len);
// printf("pfd(): ts=%u.%06u, port=%d, et=%04x, lsap=%04x, r=%d <==\n", 
//     DAGT_S(drp->ts), DAGT_US(drp->ts), p_p->port_nbr,
//     ether_type, lsap, r);  fflush(stdout);  //28
      }
   else if (t == TYPE_ATM) {
      p_p->port_nbr = drp->d.ext.flags & DF_INTERFACE;
      p_p->p_len = ntohl(drp->d.ext.wlen);  /* On-the-wire length */
      p_p->adj_type = AT_AAL5;
      rec_len = ntohs(drp->d.ext.rlen) -  /* Bytes in Dag record */
         (drp->d.ext.e.atm.pload - (Bit8 *)drp);
      ah = ntohl(drp->d.ext.e.atm.atm_hdr);
      vpvcmask = 0x0ffffff0;	 /* VPI 8 bits, VCI 16 bits */
      ah &= vpvcmask;  ah >>= 4;
      sh = &drp->d.ext.e.atm.sh;
      if ((lsap = ntohs(sh->lsap)) == 0xAAAA) {  /* SNAP packet */
         ether_type = ntohs(sh->etype);
# if MERGE_TEST == 2
          printf("%d", pi->nbr); /* */
# endif
          }
      else {
         ether_type = ntohs(sh->etype);
         }
      r = handle_pkt(p_p, ether_type, lsap, (Bit8 *)&ehdr,
         drp->d.ext.e.atm.pload, rec_len);
      }

   else if (pi->dag_type == TT_POS) {  /* TYPE_LEGACY packets .. */
# if MERGE_TEST == 2
      printf("%d", pi->nbr); /* */
# endif
      p_p->port_nbr = 0;
      p_p->p_len = ntohl(drp->d.pos.wlen);  /* On-the-wire length */
      p_p->adj_type = AT_SONET;
      r = handle_pkt(p_p, ntohs(drp->d.pos.etype), 0, (Bit8 *)&ehdr,
         drp->d.pos.pload, sizeof(drp->d.pos.pload));
      }
   else if (pi->dag_type == TT_ATM) {
      ah = ntohl(drp->d.atm.atm_hdr);
      vpvcmask = 0x0ffffff0;	 /* VPI 8 bits, VCI 16 bits */
      ah &= vpvcmask;  ah >>= 4;
      /* printf("ts=%qx (%u.%06u), atm=%06x (%08x) ",
         drp->d.atm.ts,
         DAGT_S(drp->d.atm.ts),DAGT_US(drp->d.atm.ts),
         ah, ntohl(drp->d.atm.atm_hdr)); /* */

      sh = &drp->d.atm.sh;
      if ((lsap = ntohs(sh->lsap)) == 0xAAAA) {  /* SNAP packet */
         ether_type = ntohs(sh->etype);
# if MERGE_TEST == 2
          printf("%d", pi->nbr); /* */
# endif
          }
      else {
         ether_type = ntohs(sh->etype);
         /* printf("\natm=%06x: ", 
            ah, lsap, ether_type);
         snp = (Bit8 *)&drp->d.atm.sh;
         for (j = 0; j != 24; ++j) printf(" %02x", snp[j]);
         printf("\n");  /* --- */
         /* Will need to work out what eth_type/lsap should really be! */
         }

      p_p->port_nbr = 0;
      p_p->p_len = 48;  /* Guess it's one ATM cell (!) */
      p_p->adj_type = AT_AAL5;
      r = handle_pkt(p_p, ether_type, lsap, (Bit8 *)&ehdr,
         drp->d.atm.pload, sizeof(drp->d.atm.pload));
      }
   else if (pi->dag_type == TT_ETHER) {
# if MERGE_TEST == 2
      printf("%d", pi->nbr); /* */
# endif
      p = drp->d.eth.pload;
      rec_len = sizeof(drp->d.eth.pload);
      eh = &drp->d.eth.eh;
      if ((ether_type = ntohs(eh->type)) <= 1500) {  /* 802.3 packet */
         sh = (struct snap_hdr *)p;
         if ((lsap = ntohs(sh->lsap)) == 0xAAAA) {  /* SNAP packet */
            ether_type = ntohs(sh->etype);
            p += sizeof(struct snap_hdr);
            rec_len -= sizeof(struct snap_hdr);
            }
         else ether_type = 0;  /* Other 802.2 packet */
         }
      else lsap = 0;  /* Blue book packet */

      p_p->port_nbr = 0;
      p_p->p_len = ntohs(drp->d.eth.wlen);  /* On-the-wire length */
      p_p->adj_type = AT_ETHERNET;
      r = handle_pkt(p_p, ether_type, lsap, (Bit8 *)eh, p, rec_len);
      }

   /*  for (j = 8; j != 24; ++j) printf(" %02x", drp->pload[j]);
	      printf(" ");  /* */
# if PKT_DEBUG
   if (p_p->pkt_state != 1) {
      printf("pkt_from_dag(3): not got, p_p=%x\n", p_p);  fflush(stdout);
      quack(403);
      }
# endif
// //  free_pkt(p_p);  return NULL;  // //

   if (r) {
#if 0
      if (!live_source) {
            /* Assume trace file is dagmerge of *two* Dag files */
         /* Not any more, dd_ntm can merge two trace files!
            Need a command-line option to specify 'dag_merged file' */
         p_p->Low.Interface = 1;
# if 0 
         p_p->Low.Interface = htons(
            drp->ts & 0x1ULL);  /* LSB of timestamp */
# endif
         }
#endif

# if PKT_DEBUG
   if (p_p->pkt_state != 1) {
      printf("pkt_from_dag(4): not got, p_p=%x\n", p_p);  fflush(stdout);
      quack(404);
      }
# endif
      return p_p;
      }
   else {
# if PKT_DEBUG
      if (p_p->pkt_state == 0) {
         printf("handle_dag(): already free, p_p=%x\n", p_p);  fflush(stdout);
         }
# endif
# if 0  /* Look at packet rejected by handle_pkt() /**/
      if (running_rts != 0) {
         Bit8 *snp;  int j;
         printf("h-p_r(): r=%d, ether_type=%04x (%d), lsap=%04x\n   ",
            r, ether_type,ether_type, lsap);
         snp = (Bit8 *)&drp->d.atm.sh;
         for (j = 0; j != 48; ++j) printf(" %02x", snp[j]);
         printf("\n");
         fflush(stdout);
         }
# endif
# if PKT_DEBUG > 1
      printf(" rdr: rejected, p_p=%x\n", p_p);  fflush(stdout);
# endif
      free_pkt(p_p); 
      }
   return NULL;
   }

long long dag_ts_tol = US_DAGT(0);
/* Dag cards with >1 interface can queue records on one interface
   ahead of records on another interface.  This shows up as
   backward jumps in the sequence of timestamps from the card.
   dag_ts_tol is the our allowed backward time jump in microsends */

int next_dag_record(struct interface_info *pi);

int last_dag_flags = 0;  //23

void dag_merge()
{
   long long ts, mints;   /* Dag timestamp (64-bit) */
   int j, minj, prev_badts = 0;
   int t_delta, save_lts = 0;
   struct interface_info *pi;
   struct pkt *pp;
#if MULTI
   Bit32 dag_merge_proct,
      last_dag_merge_proct = 0;
   struct tms tbuf;
#endif
   struct dag_rec *drp;
#if LOG_THREAD_PROCTIMES
   int this_tn = assign_thr_log_nbr();
   sprintf(thread_name[this_tn], "dag_merge");
   thread_proctime_org[this_tn] = thread_proctime[this_tn] = 0;
#endif   
   char buf[30], buf2[30]; //21
   long long ts_delta;
   for (;;) {
      minj = -1;
      for (j = 0; j < n_interfaces; ++j) {  /* May only be one interface */
         pi = pci[j];
         if (minj < 0) {
            mints = pi->dbp->ts;  minj = j;
            }
         else {
            ts = pi->dbp->ts;
            if (ts < mints) {
               mints = ts;  minj = j;
               }
            }
         }
      if (minj < 0) log_msg(LOG_ERR, 1, "No data on both i/fs <"); /**/
/*      printf("%d", minj);  fflush(stdout);  /**/

      pi = pci[minj];
      drp = pi->dbp;
      if (--pi->sample_count == 0) {  /* Process only 1 pkt per SampleRate */
         pi->sample_count = pi->SampleRate;
         save_lts = last_t_sec;

         if (mints < last_ts && (last_ts - mints >= dag_ts_tol)) {
               /* Backward time jump; ignore pkt */
               /* Can appear when reading from trace file */
#if 0
   ts_delta = last_ts - mints;
   fmt_ts(buf, mints);  fmt_ts(buf2, last_ts);
   printf("%s %s  -%d.%06d  %8d %10d.%06d: %02x-%02x:%02x %u %u %u  %u %u ",
	  buf2, buf, DAGT_S(ts_delta),DAGT_US(ts_delta),
      pi->ipackets, DAGT_S(pi->dbp->ts), DAGT_US(pi->dbp->ts),
	  pi->dbp->d.ext.type, pi->dbp->d.ext.flags, last_dag_flags,
      ntohs(pi->dbp->d.ext.rlen),
      ntohs(pi->dbp->d.ext.lctr), ntohs(pi->dbp->d.ext.wlen),
      pi->dbp->d.ext.e.eth.offset, pi->dbp->d.ext.e.eth.pad);
   for (j = 0; j != 8; ++j) printf(" %02x",
 	pi->dbp->d.ext.e.eth.pload[j] ); 
   printf("\n");
#endif
            ++pi->badtspkts;
            if (!live_source && !prev_badts) {
               printf(minj == 0 ? "*" : "+");  fflush(stdout);
               }
            prev_badts = 1;
            }
         else {
            pp = pkt_from_dag_record(drp, pi);
            if (pp != NULL) {  /* We have a packet */
# if PKT_DEBUG
   if (pp->pkt_state != 1) {
      printf("dag_merge(): not got, pp=%x\n", pp);  fflush(stdout);
      quack(501);
      }
# endif
               last_ts = mints;  /* Used by uptime() routines */
               t_pkt_monitor(pp);
               }
            prev_badts = 0;
            }
         }
      last_dag_flags = pi->dbp->d.ext.flags;  //23

      ++pi->ipackets;  /* For global statistics */
      ++pi->nrecords;  /* For osp() 5-min log records */
      ++pi->n_pkts;  

      if (!live_source) {  /* Reading a trace file */
         if (pi->nrecords % 10000 == 0) {  /* Indicate progress */
            printf(minj == 0 ? "." : ":");  fflush(stdout);
            }
 // if (pi->nrecords == 200) stop_meter(3);  //03
         if (!next_dag_record(pi)) {  /* EOF */
            ntm_semaphore_up("dm", &merge_sync_empty);
                /* Block until non_live_source decides what to do */
            ntm_semaphore_down("dm", &merge_sync_full);
            }
         else if (last_t_sec > save_lts) {
            /* last_t_sec updated in handle_pkt(), from t_pkt_monitor() */
            t_delta = last_t_sec - save_lts;
            if (t_delta != 1) printf("%d", t_delta);
            save_lts = last_t_sec;
            if (pi->nrecords == 1 ||  /* Just getting started */
                  (trace_count -= t_delta) <= 0) {
               ntm_semaphore_up("dm", &merge_sync_empty);
                  /* Block until non_live_source decides what to do */
               ntm_semaphore_down("dm", &merge_sync_full);
	       }
            }
         }
      else next_dag_record(pi);

# if MULTI
      if (pi->n_pkts % 1000 == 0) {
         times(&tbuf);
         dag_merge_proct = tbuf.tms_utime+tbuf.tms_stime;
         total_thread_proct += (dag_merge_proct - last_dag_merge_proct);
#  if LOG_THREAD_PROCTIMES
         thread_proctime[this_tn] += (dag_merge_proct - last_dag_merge_proct);
#  endif   
         last_dag_merge_proct = dag_merge_proct;
         }
# endif
      }
   }

int next_trace_record(struct interface_info *pi)
{

   int rl, bl, left_ib;

   if (pi->nblocks != 0) {  /* Current record is OK */
      if (pi->dag_type != TT_EXT) rl = 64;
      else  rl = ntohs(pi->dbp->d.ext.rlen);
      pi->dbp = (struct dag_rec *)( ((Bit8 *)pi->dbp) + rl );
      pi->fbx += rl;
      bl = pi->fbr - pi->fbx;
      if (bl > 16) {  /* Enough for Dag ERF header */
         if (pi->dag_type != TT_EXT) rl = 64;
         else  rl = ntohs(pi->dbp->d.ext.rlen);
         if (bl >= rl) {  /* Enough for whole record */
#if DEBUG_TF_READ > 1
            printf("%8d %10d.%06d: %02x-%02x %u %u %u\n",
               pi->ipackets, DAGT_S(pi->dbp->ts), DAGT_US(pi->dbp->ts),
	       pi->dbp->d.ext.type, pi->dbp->d.ext.flags,
	       rl, ntohs(pi->dbp->d.ext.lctr), ntohs(pi->dbp->d.ext.wlen) );
	    fflush(stdout);
#endif
            return 1;
            }
         }
      }

   left_ib = pi->fbr - pi->fbx;
   memcpy(pi->fbuf, &pi->fbuf[pi->fbx], left_ib);
#if HAVE_LIBZ
   if ((pi->fbr = gzread(pi->gfp, &pi->fbuf[left_ib], DF_BLK_SZ-left_ib))
         == 0) {
      gzclose(pi->gfp);
      pi->gfp = NULL;
#else
   if ((pi->fbr = read(pi->t_fd, &pi->fbuf[left_ib], DF_BLK_SZ-left_ib))
         == 0) {
      close(pi->t_fd);
      pi->t_fd = -1;
#endif
      return 0;  /* End of file */
      }
   ++pi->nblocks;
   pi->fbr += left_ib;
   pi->fbx = 0;
   pi->dbp = (struct dag_rec *)pi->fbuf;
#if DEBUG_TF_READ > 1
   printf("\n%8d %10d.%06d: %02x-%02x %u %u %u\n",
      pi->ipackets, DAGT_S(pi->dbp->ts), DAGT_US(pi->dbp->ts),
      pi->dbp->d.ext.type, pi->dbp->d.ext.flags,
      ntohs(pi->dbp->d.ext.rlen),
      ntohs(pi->dbp->d.ext.lctr), ntohs(pi->dbp->d.ext.wlen) );
   fflush(stdout);
#endif
   return 1;
   }

int next_dag_record(struct interface_info *pi)
{
   int rl;
   struct timespec rqt; //28

   if (!live_source) {
//     if (pi->nrecords >= 200) stop_meter(99);  // //
      return next_trace_record(pi);
      }
   else {
      if (pi->r_pkts == 0) {
         for (;;) {
            ntm_semaphore_up("ndr", &pi->db_empty);
               /* Wait while run_dag_read gets the next block */
            ntm_semaphore_down("gnb", &pi->db_ready);
	    if (pi->r_pkts != 0) break;
            rqt.tv_sec = 0;  rqt.tv_nsec = 1200000;  /* 1.2 ms */
	    /* CAUTION: nanosleep busy-waits for times < 1 ms */
            nanosleep(&rqt, NULL);
            }
         --pi->r_pkts;  /* Return first pkt in buffer */
         }
      else {
         if (pi->dag_type != TT_EXT) rl = 64;
         else  rl = ntohs(pi->dbp->d.ext.rlen);
         pi->dbp = (struct dag_rec *)( ((Bit8 *)pi->dbp) + rl );
         --pi->r_pkts;  /* pkts remaining in buffer */
         }
      }
   return 1;
   }

void *run_dag_read(void *arg)
{
   struct interface_info *pi = (struct interface_info *)arg;
   unsigned s_top, top;
   struct dag_rec *dbp;
   int rl, nip, nib;

   Bit32 dag_read_proct, last_dag_read_proct = 0;
   struct tms tbuf;
#if LOG_THREAD_PROCTIMES
   int this_tn = assign_thr_log_nbr();
   sprintf(thread_name[this_tn], "dag_read %d", pi->nbr);
   thread_proctime_org[this_tn] = thread_proctime[this_tn] = 0;
#endif   

   s_top = 0;
   for (;;) {
// printf("if %d, r_d_r() down empty\n", pi->nbr);  // //
      ntm_semaphore_down("rdr", &pi->db_empty);

#if DAG_DELAY_READ
         timeout.tv_sec = 0;  timeout.tv_usec = select_us;
         select(FD_SETSIZE, 0, 0, 0, &timeout);
#endif

      errno = 0;  /* System call error variable */
      pi->bottom = s_top;
      top = dag_offset(pi->fd, &pi->bottom, 0);
// printf("i/f %d, dag_offset: s_top=%u, bottom=%u, top=%u\n",
//   pi->nbr, s_top, pi->bottom, top); //02
//  fflush(stdout);
      dbp = (struct dag_rec *)(pi->dbuf + pi->bottom);
      s_top = pi->bottom;  /* dag_offset may change bottom,
                              top stays correct relative to bottom */
      nip = nib = 0;
      for (;;) {
         if (pi->dag_type != TT_EXT) rl = 64;
         else {
            rl = ntohs(dbp->d.ext.rlen);
            if (dbp->d.ext.lctr) {
               pi->idrops += ntohs(dbp->d.ext.lctr);
               if (dbp->d.ext.lctr == 0xFFFF) {
                  ++pi->n_stuck_lctrs;
                  log_msg(LOG_INFO, 0,
                     "i/f %d lctr=0xFFFF, can't count lost packets", pi->nbr);
                  }
               }
            }

         if (nib + rl > sizeof(struct dag_rec)*DAG_BUF_SZ) {
               /* Won't fit in in_buf */
            break;
            } 

         if (s_top + rl == top) {  /* No more data */
            break;  /* Leave one record in buffer <<<<<
               On Linux dag_offset suspendss until it gets a record,
               but on FreeBSD it seems to loop in thread-handler routines.
               Leaving a record in the buffer means that dag_offset
               always returns at least one record; next_dag_record()
               can nanosleep() when it gets 0 records, i.e. no busy-wait */
            } 

         if (s_top + rl > top) {  /* Partial record (shouldn't happen) */
            log_msg(LOG_INFO, 0,
               "i/f %d, got partial rec: %d B, rl=%d, %d pkts",
                pi->nbr, rl,nib, nip);
            break;
            }
         dbp = (struct dag_rec *)( ((Bit8 *)dbp) + rl );
         s_top += rl;
         nib += rl;  ++nip;
         }

      pi->dbp = (struct dag_rec *)(pi->dbuf + pi->bottom);
      pi->r_pkts = nip;
      pi->card_ts = pi->dbp->ts;
      ++pi->nblocks;
// if (nip != 0)  //02
//    printf("if %d, r_d_r() %d pkts, bottom=%u, s_top=%u, top=%u\n",
//       pi->nbr, pi->r_pkts, pi->bottom, s_top, top); //02 
      ntm_semaphore_up("rdr", &pi->db_ready);
 
      times(&tbuf);
#if MULTI
      dag_read_proct = tbuf.tms_utime+tbuf.tms_stime;
      total_thread_proct += (dag_read_proct - last_dag_read_proct);
# if LOG_THREAD_PROCTIMES
      thread_proctime[this_tn] += (dag_read_proct - last_dag_read_proct);
# endif   
      last_dag_read_proct = dag_read_proct;
#endif
      }
   }

void init_live_sources(void)
{
   int j;
   struct interface_info *pi;
   for (j = 0; j != n_interfaces; ++j) {
      pi = pci[j];
      while (pi->nblocks == 0) next_dag_record(pi);
      pi->running = 1;
      if (pi->dbp->d.ext.type == 0 && pi->dag_type == TT_EXT)
         log_msg(LOG_ERR, 1,
            "Trace file %s has LEGACY records; use -g to specify Dag type!", pi->name);
      }

   meter_start_time = pci[0]->card_ts;  /* Set to min Dag timestamp */
   for (j = 1; j != n_interfaces; ++j) {
      if (pci[j]->card_ts <= meter_start_time)
         meter_start_time = pci[j]->card_ts;
      }

   log_msg(LOG_INFO, 0, "init_live_sources(): first blocks read");
   fflush(stdout);
   }

# if !DAG_NEW_API  /* Old version */

void bad_ts_msg(char *msg, struct interface_info *pi,
   int n, int ni, long long ts, long long tb, char *comment)
   /* ts = last good timestamp, tb = timestamp from buffer.
      diff (as printed) = tb - ts */
{
   int d = DAGT_S(tb - ts);
   char b_ts[30], b_tb[30];
   fmt_ts(b_ts, ts);  fmt_ts(b_tb, tb);
   log_msg(LOG_ERR, 0, "%s i/f %d: bytes=%u/%u, %s-%s, diff=%ds%s",
      msg, pi->nbr, n,ni, b_ts,b_tb, d, comment);
   }

int ts_seq_ok(long long ts, long long tb)
{
   long long td = tb-ts;
   if (td < 0) return 0;  /* Timestamps must be increasing */
   return DAGT_S(td) <= MX_TS_JUMP;
   }

void dag_merge()
{
   long long ts, mints;   /* Dag timestamp (64-bit) */
   int j, minj, save_lts, n, rec_len;
   struct interface_info *pi;
   struct pkt *pp;
#if MULTI
   Bit32 dag_merge_proct,
      last_dag_merge_proct = 0;
   struct tms tbuf;
#endif
#if LOG_THREAD_PROCTIMES
   int this_tn = assign_thr_log_nbr();
   sprintf(thread_name[this_tn], "dag_merge");
   thread_proctime_org[this_tn] = thread_proctime[this_tn] = 0;
#endif   

   for (;;) {
      minj = -1;
      for (j = 0; j < n_interfaces; ++j) {  /* May only be one interface */
         pi = pci[j];
         if (pi->r_bytes == 0) {  /* Empty buffer */
            pi->n_pkts = 0;
#if MULTI
	    update_hist(&dm_hist, 'r',j);
            ntm_semaphore_down("gnb", &pi->db_ready);
            pi->drp = (struct dag_rec *)pi->in_buf[pi->outx];
            pi->r_bytes = pi->in_bytes[pi->outx];
/*          printf("\n++++ got buffer %d /%d, %d pkts\n",
               pi->nbr,pi->outx, pi->r_bytes);  /**/
#else
            waiting_live_source = j;
            return;  /* Keep outer block alive */
#endif
            }
         if (minj < 0) {
            mints = pi->drp->ts;  minj = j;
            }
         else {
            ts = pi->drp->ts;
            if (ts < mints) {
               mints = ts;  minj = j;
               }
            }
         }
      if (minj < 0) log_msg(LOG_ERR, 1, "No data on both i/fs <"); /**/
/*      printf("%d", minj);  fflush(stdout);  /**/

      pi = pci[minj];
      last_ts = mints;  /* Used by uptime() routines */
      if (pi->dag_type != TT_EXT) rec_len = 64;  /* Legacy, fixed-len */
      else {
         rec_len = ntohs(pi->drp->d.ext.rlen);
#if 0
   if (rec_len != pi->dag_slen + 16)  /* 16 bytes Dag header before payload */
      printf("short (varlen) header, rec_len=%d, dag_slen=%d\n",
          rec_len, pi->dag_slen);
#endif
         }
      if (--pi->sample_count == 0) {  /* Process only 1 pkt per SampleRate */
         pi->sample_count = pi->SampleRate;
         save_lts = last_t_sec;

         pp = pkt_from_dag_record(pi->drp, pi);
	 if (pp != NULL) {  /* We have a packet */
# if PKT_DEBUG
   if (pp->pkt_state != 1) {
      printf("dag_merge(): not got, pp=%x\n", pp);  fflush(stdout);
      quack(501);
      }
# endif
            t_pkt_monitor(pp);
            }
         if (!live_source &&  /* Reading a trace file */
               last_t_sec > save_lts) {
            if (--trace_count == 0) {
               if (++pi->outx == N_DAG_BUFS) pi->outx = 0;
               return;  /* Keep trace file loop alive */
               }
            }
         }
      ++pi->ipackets;  /* For global statistics */
      ++pi->nrecords;  /* For osp() 5-min log records */
      ++pi->n_pkts;  
      if ((pi->r_bytes -= rec_len) == 0) {  /* End of buffer */
/*       printf("\n---- end buffer %d/%d, %d pkts\n",
            pi->nbr,pi->outx, pi->npkts);  /**/
# if MULTI
         if (++pi->outx == N_DAG_BUFS) pi->outx = 0;
         update_hist(&dm_hist, 'e',minj);
         ntm_semaphore_up("gnb", &pi->db_empty);

         times(&tbuf);
         dag_merge_proct = tbuf.tms_utime+tbuf.tms_stime;
         total_thread_proct += (dag_merge_proct - last_dag_merge_proct);
#  if LOG_THREAD_PROCTIMES
         thread_proctime[this_tn] += (dag_merge_proct - last_dag_merge_proct);
#  endif   
         last_dag_merge_proct = dag_merge_proct;
# endif
         }
      else {
         pi->drp = (struct dag_rec *)( ((Bit8 *)pi->drp) + rec_len );
# if DAG_EXT_TESTING
         if (pi->drp->d.ext.type != TYPE_HDLC_POS) {
            log_msg(LOG_INFO, 0, "dag_merge(1): didn't get POS type");
   quack(111);
            }
# endif
         ts = pi->drp->ts;  /* Look at next record */
         if (!ts_seq_ok(mints, ts)) {  /* Look one record further on */
            if (pi->dag_type != TT_EXT) rec_len = 64;
            else rec_len = ntohs(pi->drp->d.ext.rlen);
            pi->drp = (struct dag_rec *)( ((Bit8 *)pi->drp) + rec_len );
            pi->r_bytes -= rec_len;
    quack(99);
# if DAG_EXT_TESTING
            if (pi->drp->d.ext.type != TYPE_HDLC_POS) {
               log_msg(LOG_INFO, 0, "dag_merge(2): didn't get POS type");
   quack(222);
               }
# endif
            ++pi->idrops;
            if (pi->r_bytes >= 8 && ts_seq_ok(last_ts, pi->drp->ts)) {
               bad_ts_msg("*?*",
                  pi, pi->r_bytes, pi->in_bytes[pi->outx], last_ts, ts, 
                  " single bad timestamp, ignored");
               continue;
               }
            bad_ts_msg("*?*",
               pi, pi->r_bytes, pi->in_bytes[pi->outx], mints, ts, 
               " rest of buffer abandoned");
            for ( ; pi->r_bytes != 0; pi->r_bytes -= rec_len) {
# if DAG_EXT_TESTING
               if (pi->r_bytes < 0) {
                  log_msg(LOG_INFO), 0, "dag_merge(3): negative bytes left!");
	quack(333);
                  }
# endif
               if (pi->dag_type != TT_EXT) rec_len = 64;
               else rec_len = ntohs(pi->drp->d.ext.rlen);
               pi->drp = (struct dag_rec *)( ((Bit8 *)pi->drp) + rec_len );
               ++pi->idrops;
# if DAG_EXT_TESTING
               if (pi->drp->d.ext.type != TYPE_HDLC_POS) {
                  log_msg(LOG_INFO, 0, "dag_merge(4): didn't get POS type");
   quack(444);
                  }
# endif
               }
# if MULTI
            if (++pi->outx == N_DAG_BUFS) pi->outx = 0;
            update_hist(&dm_hist, 'b',minj);
            ntm_semaphore_up("gnb", &pi->db_empty);

            times(&tbuf);
            dag_merge_proct = tbuf.tms_utime+tbuf.tms_stime;
            total_thread_proct += (dag_merge_proct - last_dag_merge_proct);
#  if LOG_THREAD_PROCTIMES
            thread_proctime[this_tn] += (dag_merge_proct - last_dag_merge_proct);
#  endif   
            last_dag_merge_proct = dag_merge_proct;
# endif
            }
         }
      }
   }

FILE *dag_data_log = NULL;

void log_dag_data(int cmd, int iface, int block,
   struct dag_rec *d_buf, int bytes)
{
   Bit8 *bp;
   int recs, j, k;

   if (dag_data_log == NULL)
      dag_data_log = fopen("dag_data", "w");

   switch (cmd) {
   case 1:  /* Log last 2 records plus trailing bytes */
      fprintf(dag_data_log, "i/f %d, block %d, last 2+ records:\n",
         iface, block);
      recs = (bytes + sizeof(struct dag_rec)-1)/sizeof(struct dag_rec);
      j = (recs-3)*sizeof(struct dag_rec);
      bp = (Bit8 *)d_buf + j;
      break;
   case 2:   /* Log first 3 records */
      fprintf(dag_data_log, "i/f %d, block %d, first 3 records:\n", iface, block);
      j = 0;
      bp = (Bit8 *)d_buf;
      break;
      }   
   for (k = 0; k != 3; ++k) {
      fprintf(dag_data_log, "   %02x", *bp++);
      for (++j; j < bytes; ) {
         fprintf(dag_data_log, " %02x", *bp++);
         if (++j % sizeof(struct dag_rec) == 0) break;
         if (j % 32 == 0) fprintf(dag_data_log, " ");
         }
      fprintf(dag_data_log, "\n");
      }
   fprintf(dag_data_log, "\n");
   }

# if DAG_HW_RESET
void dag_reset(int i)
{  
  /* Sometimes, after running happily for hours or days, a Dag card
     can just stop writing cells into memory.  When this happens,
     DAG_DIRECT reads just keep returning 0 bytes + errno 11 (EAGAIN).
     A level 2 reset of the Dag card seems to get it working again!
        Nevil (with help from Stephen and David), Fri 11 Jan 02 */

   dagioaddr dagioa;
   int r;

   /* The card has a reset register at 0x88.  The relevant bitmap is:
         Bit 31: Trigger level 2 reset (auto clearing)
         Bit 30: Trigger level 1 reset (not auto clearing) */

   dagioa.offset = 0x88;  dagioa.val = 0x80000000;
   r = ioctl(pci[i]->fd, DAG_IOWRITE, &dagioa);

   log_msg(LOG_ERR, 0,"=== dag_reset(%d): ioctl returned %d", i,r);
   if (r != 0)
      log_msg(LOG_ERR, 0,"   error %d, %s", errno, strerror(errno));
   }
# endif

int read_live_block(int i)  /* DAG_DIRECT version */
{
   struct interface_info *pi = pci[i];
   int r, en, s, nb, rr;
   long long read_ts = 0LL;
   char g_ts[30], b_ts[30];
   Bit8 *p8;
   long long d;
   struct dag_rec *drp;

   struct pkt *pp, *last_pp;
   long long last_ts, ts;

   r = read(pi->fd, pi->in_buf[pi->inx],
         sizeof(struct dag_rec)*DAG_BUF_SZ);
   if (got_USR1) {
      printf(" %d:%d", pi->nbr,r);
      fflush(stdout);
      }
#  if IF_READ_TRACE
   printf(" %d:%d %d", pi->nbr,r, pi->nblocks);
#  endif
   if (r <= 0) {
      if (!live_source) {
         if (r == 0) {  /* EOF on trace file */
            show_memory();
            return 0;
            }
         }
      else if ((en = errno) != pi->last_errno) {
         if (en != EAGAIN) log_msg(LOG_ERR, 0,
            "if_read(1): i/f %d, error %d (%s)", pi->nbr, en, strerror(en));
         pi->last_errno = errno;
         }

      /* No data.  This includes EAGAIN reads! */
      if (pi->empty > MX_NODATA_READS) {  /* Guess lowest ts for next pkt */
          pi->card_ts = last_ts + select_time;
         if (!pi->no_data_warned) {
            fmt_ts(g_ts, pi->card_ts);
            log_msg(LOG_WARNING, 0, "i/f %d, no data, ts set to %s",
               pi->nbr, g_ts);
            pi->no_data_warned = 1;
            }
#  if DAG_HW_RESET
         dag_reset(i);  /* Try to reset the Dag card */
#  else
         log_msg(LOG_ERR, 211, "Don't bother to restart: shutting down");
#  endif
        } 

      r = 0;
      s = S_EMPTY;  /* No data in pipe */
      }
   else {  /* Data bytes read */
# if MERGE_TEST == 2
      printf("\nif_read(2): i/f %d, %d records read ",
         pi->nbr, r/sizeof(struct dag_rec));
# endif
      pi->no_data_warned = pi->last_errno = 0;
      rr = r;

      drp = (struct dag_rec *)pi->in_buf[pi->inx];
      read_ts = drp->ts;
# if IF_READ_TRACE
      fmt_ts(b_ts, read_ts);
      printf(" %d %s-", rr,b_ts);
# endif
      if (r != 0) {
         if (pi->nblocks == 0) {  /* Just getting started */
            mutex_lock("rlb", &start_time_mutex);
            if (meter_start_time == 0 || read_ts < meter_start_time)
               pi->card_ts = meter_start_time = read_ts;
            last_ts = meter_start_time;
            last_t_sec = DAGT_S(last_ts);
            mutex_unlock("rlb", &start_time_mutex);
            }

         if ((pi->running || pi->nblocks != 0) &&  /* We've read some data */
              !ts_seq_ok(pi->card_ts, read_ts)) {
                  /* Bad timestamp diff between blocks */
            r = 0;
            d = read_ts - pi->last_read_ts;
            if (d > 0) {
               pi->idrops += nip;  /* Records lost */
               fmt_ts(g_ts, pi->card_ts);  fmt_ts(b_ts, read_ts);
               if (DAGT_S(d) <= MX_TS_JUMP*10) {  /* Tail-biting */
                  log_msg(LOG_ERR, 0,
                     "if_read(5a): i/f %d, tail-biting (last=%s, this=%s)."
                     "  Reset card_ts", pi->nbr, g_ts,b_ts);
                  pi->idrops += DAG_BUF_SZ;  /* Vague guess! */
                     /* We restart the merge from this timestamp, so
                        we loose the whole 32 MB in the buffer ring */
                  r = rr;  /* Carry on from this block */
                  }
               else {  /* Garbage block */ 
                  log_msg(LOG_ERR, 0,
                     "if_read(5b): i/f %d, %d bytes ignored"
                     " (last=%s, this=%s)", pi->nbr, pi->r_bytes, g_ts,b_ts);
                  pi->idrops += nip;
                  pi->last_read_ts = read_ts;
                  }
               }
	    else log_msg(LOG_ERR, 0, "Backward ts jump, d = %lld <<<", d); 
            }

         if (r != 0) {  /* Timestamps OK between blocks */
            pi->in_bytes[pi->inx] = r;
            pi->card_ts = read_ts;
            dbp = (struct dag_rec *)(pi->in_buf[pi->inx] + l_nib);
            pi->card_ts = dbp->ts;  /* Last record in block */
# if !MULTI
            pi->drp = pi->in_buf[0];
            pi->r_bytes = pi->in_bytes[0];
# endif
# if IF_READ_TRACE
            fmt_ts(b_ts, pi->card_ts);
            printf(" %s %d\n", b_ts, nrecs);  fflush(stdout);
# endif
            pi->last_read_ts = read_ts;
            }

         if (r == 0) {  /* Couldn't recover */
            ++pi->badtspkts;
            if (++pi->consec_badts == MX_CONSEC_BADTS) {
               log_msg(LOG_ERR, 0,
                  "Too many bad timestamps, try to restart i/f %d",
                  pi->nbr);
               if (pi->restarts) log_msg(LOG_ERR, 111,
                  "No data after previous restart, shutting down");
               else {
#if DAG_HW_RESET
                  dag_reset(i);  /* Try to reset the Dag card */
#else
                  log_msg(LOG_ERR, 211, "Don't try to restart: shutting down");
#endif
                  }
               }
            else if (pi->consec_badts > MX_CONSEC_BADTS)
               log_msg(LOG_ERR, 113, "Restart insufficient, shutting down");
            }
         else pi->consec_badts = 0;
         }

      if (r != 0) ++pi->nblocks;

      s = rr == DAG_BUF_SZ*sizeof(struct dag_rec) ? S_FULL : S_NORMAL;
#if MERGE_TEST >= 1
      fmt_ts(b_ts, pi->card_ts);
      printf("\nif_read(7): i/f %d, ts=%s, %d ccells, %d nblocks\n", 
         pi->nbr, b_ts, nrecs, pi->nblocks);
#endif
      }
   
   switch (pi->last_s) {
   case S_NORMAL:
      switch (s) {
      case S_NORMAL:
         break;
      case S_FULL:
         pi->full = 1;
         break;
      case S_EMPTY:
         pi->empty = 1;
         break;
         }
      break;
   case S_FULL:
      switch (s) {
      case S_NORMAL:
         if (pi->full > pi->mx_full) pi->mx_full = pi->full;
         pi->full = 0;  
         break;
      case S_EMPTY:
         if (pi->full > pi->mx_full) pi->mx_full = pi->full;
         pi->full = 0;  pi->empty = 1;
         break;
      case S_FULL:
         ++pi->full;
         break;
         }
      break;
   case S_EMPTY:
      switch (s) {
      case S_NORMAL:
         if (pi->empty > pi->mx_empty) pi->mx_empty = pi->empty;
         pi->empty = 0;  
         break;
      case S_FULL:
         if (pi->empty > pi->mx_empty) pi->mx_empty = pi->empty;
         pi->empty = 0;  pi->full = 1;
         break;
      case S_EMPTY:
         ++pi->empty;
         break;
         }
      break;
      }
   pi->last_s = s;
   return r;
   }

void *run_dag_read(void *arg)
{
   struct interface_info *pi = (struct interface_info *)arg;
   int r;
   struct timeval timeout;
   Bit32 dag_read_proct,
      last_dag_read_proct = 0;
   struct tms tbuf;
#if LOG_THREAD_PROCTIMES
   int this_tn = assign_thr_log_nbr();
   sprintf(thread_name[this_tn], "dag_read %d", pi->nbr);
   thread_proctime_org[this_tn] = thread_proctime[this_tn] = 0;
#endif   

   for (;;) {
      ntm_semaphore_down("rdr", &pi->db_empty);
      for (;;) {
#if DAG_DELAY_READ
         timeout.tv_sec = 0;  timeout.tv_usec = select_us;
         select(FD_SETSIZE, 0, 0, 0, &timeout);
#endif
         if (read_live_block(pi->nbr - 1) != 0) break;
         }
      if (++pi->inx == N_DAG_BUFS) pi->inx = 0;
      ntm_semaphore_up("rdr", &pi->db_ready);
 
      times(&tbuf);
      dag_read_proct = tbuf.tms_utime+tbuf.tms_stime;
      total_thread_proct += (dag_read_proct - last_dag_read_proct);
#if LOG_THREAD_PROCTIMES
      thread_proctime[this_tn] += (dag_read_proct - last_dag_read_proct);
#endif   
      last_dag_read_proct = dag_read_proct;
     }
   }

void init_live_sources(void)
{
#define STARTUP_TRIES  50
   int j,n,b;
   for (n = 0; n != STARTUP_TRIES; ++n) {
      for (j = 0; j != n_interfaces; ++j) {  /* Get block for each i/f */
         read_live_block(j);
         }
      for (b = j = 0; j != n_interfaces; ++j)
         if (pci[j]->nblocks != 0) ++b;
      if (b == n_interfaces) break;
      }

   if (n == STARTUP_TRIES) log_msg(LOG_ERR, 98,
      "init_live_sources(): couldn't get data from interface(s)");
#if !MULTI
   meter_start_time = pci[0]->card_ts;  /* Set to min Dag timestamp */
   for (j = 1; j != n_interfaces; ++j) {
      if (pci[j]->card_ts <= meter_start_time)
         meter_start_time = pci[j]->card_ts;
      }
   last_ts = meter_start_time;
   last_t_sec = DAGT_S(last_ts);
#else
   for (n = 0; n != N_PME_BUFS; ++n) {  /* Discard current Dag buffers */
      for (j = 0; j != n_interfaces; ++j) {
         read_live_block(j);
         }
      }
   for (j = 0; j != n_interfaces; ++j) {
      pci[j]->running = 1;  /* OK to build chains of pkts */
      for (n = 0; n != N_DAG_BUFS; ++n)  /* Now let run_dag_read() fill buffers */
         ntm_semaphore_up("init", &pci[j]->db_empty);
      }
#endif
   printf("init_live_sources(): first blocks read\n");
   fflush(stdout);
   }
# endif  /* !DAG_NEW_API */

int open_dag_direct(struct interface_info *pi)
{
   int ifd, flags;
   char options[50], *vopt;

   if (!live_source) {
#if HAVE_LIBZ
      if ((pi->gfp = gzopen(pi->name,"r")) == NULL) {
#else
      if ((pi->t_fd = open(pi->name, O_RDONLY)) < 0) {
#endif
         log_msg(LOG_ERR, 0,
            "i/f open(%s) failed: %s\n", pi->name, strerror(errno));
         return 0;  /* Fail */
         }
      log_msg(LOG_WARNING, 0, "Trace file %s opened", pi->name);
      }
   else {  /* Live source */
#if DAG_NEW_API
      if (pi->running) {
         if (dag_stop(pi->fd) < 0)
            log_msg(LOG_ERR, 0, "dag_stop %s: %s\n", pi->name, strerror(errno));
         if (dag_start(pi->fd) < 0)
            log_msg(LOG_ERR, 1, "dag_start %s: %s\n", pi->name, strerror(errno));
         }
      else {
         if ((ifd = dag_open(pi->name)) < 0)
            log_msg(LOG_ERR, 1, "dag_open %s: %s\n", pi->name, strerror(errno));

	 vopt = pi->dag_varlen ? "varlen" : "fixed";
// //         sprintf(options, "slen=%d %s", pi->dag_slen, vopt);
         sprintf(options, "%s slen=%d", vopt, pi->dag_slen);
         log_msg(LOG_WARNING, 0, "Starting i/f %d, %s: options='%s'",
            pi->nbr, pi->name, options);
         if (dag_configure(ifd, options) < 0)
            log_msg(LOG_ERR, 1, "dag_configure %s: %s\n", pi->name, strerror(errno));

         if ((pi->dbuf = dag_mmap(ifd)) == MAP_FAILED)
            log_msg(LOG_ERR, 1, "dag_mmap %s: %s\n", pi->name, strerror(errno));

         if (dag_start(ifd) < 0)
            log_msg(LOG_ERR, 1, "dag_start %s: %s\n", pi->name, strerror(errno));
         }
#else
      if (pi->fd != 0) {
         if (close(pi->fd) < 0) {
            log_msg(LOG_ERR, 0,
               "i/f: close(%s) failed: %s\n", pi->name, strerror(errno));
            return 0;  /* Fail */
            }
         pi->fd = 0;
         }
      if ((ifd = open(pi->name, O_RDONLY)) < 0) {
         log_msg(LOG_ERR, 0,
            "i/f: open(%s) failed: %s\n", pi->name, strerror(errno));
         return 0;  /* Fail */
         }

      if ((flags = fcntl(ifd, F_GETFL, 0)) < 0) {
         log_msg(LOG_ERR, 0, "iif: getfl(%s): fcntl get failed\n", pi->name);
         return 0;  /* Fail */
         }
# if MULTI
      flags &= !O_NONBLOCK;  /* Blocking reads for threaded meter */
# else
      flags |= O_NONBLOCK;  /* Non-blocking reads for single-thread meter */
# endif
      if (fcntl(ifd, F_SETFL, flags) < 0) {
         log_msg(LOG_ERR, 0, "iif: nonblck(%s): fcntl set failed\n", pi->name);
         return 0;  /* Fail */
         }
#endif
      }

   pi->fd = ifd;
   return 1;  /* OK */
   }

#if DAG_NEW_API
void dag_close_all(void)
{
   int j;
   for (j = 0; j != n_interfaces; ++j) {
      if (dag_stop(pci[j]->fd) < 0)
         log_msg(LOG_ERR, 0, "dag_stop %s: %s\n", pci[j]->name, strerror(errno));
      if (dag_close(pci[j]->fd) < 0)
         log_msg(LOG_ERR, 0, "dag_close %s: %s\n", pci[j]->name, strerror(errno));
      }
   }
#   endif

int init_interface(struct interface_info *pi)  /* Dag direct */
{
   char buf[150], *crl_info_string;
   int len, ifd, flags, j;

   if (strncmp(pi->name, "/dev/", 5) != 0) {
      live_source = 0;  /* User wants to read trace file(s) */
      trace_count = trace_interval;
      crl_n_count = 0;

      sprintf(buf, "%d  -i%s (Dag trace file)",
         trace_interval,  /* NeMaC uses this for its collection interval */
         pi->name);
      len = strlen(buf);
      crl_info_string = (char *)malloc(len+1);
      strcpy(crl_info_string, buf);
      crl_info = crl_info_string;
         /* NeMaC writes this to a #CoralReef record */

      if ((pi->fbuf = malloc(DF_BLK_SZ)) == NULL) {
         log_msg(LOG_ERR, 99,
            "Not enough memory for file %s buffer!", pi->name);
         }
      }

# if !DAG_NEW_API
   for (j = 0; j != N_DAG_BUFS; ++j) {
      if ((pi->in_buf[j] = malloc(
            sizeof(struct dag_rec)*DAG_BUF_SZ)) == NULL) {
         close(pi->fd);
         log_msg(LOG_ERR, 99,
            "Not enough memory for i/f %s buffer %d!", pi->name, j);
         }
      }
   pi->inx = pi->outx = 0;
# endif /* !DAG_NEW_API */

   pi->last_errno = 0;
   if (!open_dag_direct(pi)) return 0;

#if MULTI
   if (live_source) {
      ntm_semaphore_init("dag_in_ready",pi->nbr, &pi->db_ready);
      semaphore_decrement(&pi->db_ready);  /* Thread blocked */
      ntm_semaphore_init("dag_in_empty",pi->nbr, &pi->db_empty);
      semaphore_decrement(&pi->db_empty);
      if (gprof_pthread_create(&pi->dag_thread, NULL, run_dag_read, pi)) {
         log_msg(LOG_ERR, 1, "Error creating i/f %d thread", j+1);
         }
      }
   /* Non-live source: dag_merge() does file reads in next_trace_record() */
#endif

   return 1;  /* OK */
   }

unsigned int pkt_counts(int reset)
{
   int j;
   struct interface_info *pi;

   pkts = drops = 0L;  /* Get total packets seen and dropped */
   for (j = 0; j != n_interfaces; ++j) {
      pi = pci[j];
      pkts += pi->ipackets;  drops += pi->idrops;
      }
   if (reset) {  /* Was doing this before setting pkts  19 Nov 03 */
      npackets_org = pkts;  lostpackets_org = drops;
      }
   npackets = pkts - npackets_org;
   lostpackets = drops - lostpackets_org;
   return 1;
   } 

/* ===== End of Dag direct interface code ===== */

#endif  /* !NETFLOW */


unsigned long iface_info(int if_nbr, int which)
{
   struct interface_info *pi;
   pi = pci[if_nbr-1];
   if (which == IFACE_LOST) 
      return pi->idrops + pi->noflowpackets;
   else if (which == IFACE_RATE)
      return pi->SampleRate;
   else return 0;
   }

void bump_noflowpkts(int if_nbr)
{
   ++noflowpackets;
   ++pci[if_nbr-1]->noflowpackets;  /* 1-org */
   }


void start_uptime_clock(void)
{
#if !DAG_DIRECT  /* Dag code sets meter_start_time directly */
   struct timeval tv;
# if CR_DATA
   tv = last_ts;  /* Use time from trace file */
# else
   gettimeofday(&tv, (struct timezone *)0);
# endif
   meter_start_time = tv;
   last_t_sec = 0;
/* printf("66 start_uptime_clock(): tv_sec=%u\n",
      meter_start_time.tv_sec);  /**/
#endif
   }

Bit32 uptime_s(void) /* seconds*/
{
#if NETFLOW || LFAP_METER  /* These use time() */
   struct timeval tv;
   gettimeofday(&tv, (struct timezone *)0);
   return tv.tv_sec - meter_start_time.tv_sec;
#elif DAG_DIRECT
   long long tsec = last_ts - meter_start_time;
   if (last_ts == 0 || meter_start_time == 0) return 0;
   else return DAGT_S(tsec);  
# else  /* Others use pkt timestamps from interface */
   if (last_t_sec == 0 || meter_start_time.tv_sec == 0) return 0;
   else return last_t_sec - meter_start_time.tv_sec;
#endif
   }

Bit32 uptime_cs(void) /* centiseconds */
{
#if DAG_DIRECT
   long long t = last_ts - meter_start_time;
   return DAGT_CS(t);
#else
   struct timeval tv;
# if CR_DATA
   tv = last_ts;  /* Use time from trace file */
# else
   gettimeofday(&tv, (struct timezone *)0);
# endif
   tv_sub(tv, meter_start_time);
   return tv.tv_sec*100 + tv.tv_usec/10000;
#endif
   }

Bit32 seconds(ts_type nt)
{
#if DAG_DIRECT
   long long t = nt - meter_start_time;
   return DAGT_CS(t);
#else
   tv_sub(nt, meter_start_time);
   return nt.tv_sec*100 + (Bit32)(((double)nt.tv_usec)/10000);
#endif
   }

#if DAG_DIRECT
Bit32 centiseconds(long long nt)
{
   long long t = nt - meter_start_time;
   return DAGT_CS(t);
#else
double centiseconds(struct timeval nt)
{
   tv_sub(nt, meter_start_time);
   return ((double)nt.tv_sec)*100.0 + ((double)nt.tv_usec)/10000.0;
#endif
   }

#if DAG_DIRECT
double microseconds(long long nt)
{
   long long t = nt - meter_start_time;
   return ((double)DAGT_S(t))*1000000.0 + (double)DAGT_US(t);
#else
double microseconds(struct timeval nt)
{
   tv_sub(nt, meter_start_time);
   return ((double)nt.tv_sec)*1000000.0 + (double)nt.tv_usec;
#endif
   }

struct timeval unix_ts(double us)
{
   struct timeval ut;
   ut.tv_sec = us/1000000.0;
   ut.tv_usec = us - ut.tv_sec*1000000.0;
   return ut;
   }

void show_meter_time()
{
   char msg[60], *ts;
   ts = ctime((time_t *)&meter_start_time);
   sprintf(msg,"%lu seconds since %c%c%c%c:%c%c",
      uptime_s(),
      ts[11],ts[12],ts[14],ts[15], ts[17],ts[18]);
   display_msg(0,msg);
   }

extern int errno;
int snmp_dump_packet = 0;
#define MXINTERFACES 4

char interface_name[SZ_IFNAME*MXINTERFACES+1] = "";
   /* Interface name (used by met_vars.c) */
char r_community[SZ_IFNAME+1], w_community[SZ_IFNAME+1];

/* Help screen.  Add any new command line options here */
void print_help(char *prog)
{
   fprintf(stderr, "\n%s\n\n", version_descr);
   fprintf(stderr, "Usage: %s [OPTION] ..\n\n", prog);
   fprintf(stderr, 
      "  -i IFN \t Specify interface to monitor (maximum of 4)\n"
      "  -k     \t Disable keyboard input processing\n"
      "  -l     \t Use length field from IP headers\n"
      "  -m PRT \t Use port PRT to communicate with readers/managers\n"
#if NF_OCX_BGP
      "  -o     \t Use 'owner' ASNs instead of 'next-hop'\n"
#endif
#ifdef LFAP_METER
      "  -p1    \t Display flow info received from a CCE\n"
      "  -p2    \t Display messages sent between CCE and FAS (LfapMet)\n"
      "  -p4    \t Display raw messages (hexdump) received from CCE\n"
      "  -p8    \t log flow info received from a CCE in file "
                   "'LfapMet.logflowstats'\n"
      "  -p16   \t log messages sent between CCE and FAS in file "
                   "'LfapMet.logmessages'\n"
      "  -p32   \t log raw messages (hexdump) received from CCE in file"
                   "'LfapMet.loghexmessages' \n\t\t (useful for debugging)\n"
      "         \t -p options can be combined, e.g:\n"
      "         \t -p63 enables all\n"
      "         \t -p9 enables display & logging of flow info\n"
      "  -R RSC \t Set snmp read community for CCEs\n"
#endif
      "  -r RSC \t Set snmp read community to RSC\n"
      "  -s     \t Disable screen display\n"
      "  -w WSC \t Set snmp write community to WSC\n"
      "  -D     \t Run NeTraMet as a daemon (superset of -k -s)\n"
      "  -M LFN \t Set meter log file name to LFN\n"
      "\n     \t\t Memory allocation (defaults)\n"
      "  -f fff \t fff flows (%lu)\n"
      "  -u rrr \t rrr rUles (%lu)\n"
      "  -a aaa \t aaa pAcket data blocks (%lu)\n"
      "  -t ttt \t ttt IP sTreams (%lu)\n"
      "  -b bbb \t bbb stream data Blocks (%lu)\n"
      "  -v ddd \t ddd distributions (%lu)\n"
      "  -e eee \t eee Events (%lu)\n"
      "  -O ooo \t eee OAM data records (%lu)\n"
      "  -P ppp \t ppp Pkt structs (%lu)\n"
#if DAG_DIRECT
      "\n     \t\t DAG card options (defaults)\n"
      "  -g ttt \t card type, 0=ATM, 1=Ether, 2=PoSm, 3=Ext (3)\n"
      "  -I ppp \t ports per card (1)\n"
      "  -L nnn \t bytes captured, including link-layer headers (%lu)\n"
      "  -F     \t Fixed-length capture records (default)\n"
      "  -V     \t Variable-length capture records\n"
#endif
#if USER_INFO
      "  -U uuu \t uuu Users in userinfo table (%lu)\n"
#endif
      "\n"
      "A meter reader can collect flow data from many meters, and each meter\n"
      "may have its data retrieved by serveral meter readers.  Traffic flows\n"
      "of interest are defined by user-specified sets of rules\n"
      "\n"
      "For more information see the NeTraMet Reference Manual and User Guide\n"
      "\n"
      "Report bugs to netramet@list.auckland.ac.nz\n\n",
      mxflows, mxrules, mxstr_blocks, mxstreams, 
	   mxdistribs, mxdistevents, mxpktdatas, mxoamrecs, mxpkts
#if DAG_DIRECT
      , DAG_DEFAULT_SLEN
#endif
#if USER_INFO
      , mxusers
#endif
      );
   }

void check_struct_sizes(void)
{
   /* Using Bit16 and Bit32 instead of arrays of Bit8 requires
      us to keep these things correctly aligned in memory.
      gcc pads structures to multiples of 4 bytes, but other
      compilers might not, hence this check */
#if 1
   struct stream s;
# if MULTI
   printf("size(mutex)=%d\n\n", sizeof(pthread_mutex_t));
# endif
# if DAG_DIRECT
   printf("size(dag_rec) = %d\n", sizeof(struct dag_rec));
   if (sizeof(struct dag_rec) != 64)
      printf("   size(dag_rec) should be 64 bytes <<<<<\n");
   printf("\n");
# endif
   printf("size(struct pkt):\n"
      "   address=%d, pkt_key=%d, pkt=%d\n",
      sizeof(union address), sizeof(struct pkt_key), sizeof(struct pkt));
   printf("size(flowhash.h):\n"
      "   stream_name=%d, key=%d, flow_key=%d, rule=%d\n",
      sizeof(struct stream_name), sizeof(struct key),
      sizeof(struct flow_key), sizeof(struct rule));

   printf("   struct stream: offset of sfn = %d\n",
      (IntFromPtr)&s.sfn - (IntFromPtr)&s.next_hc);

   if (sizeof(struct pkt_key) % 4 != 0 ||
         sizeof(struct stream_name) % 4 != 0 ||
         sizeof(struct key) % 4 != 0 ||
         sizeof(struct flow_key) % 4 != 0 ) {
      printf("sizeof(struct) not a multiple of 4 bytes <<<<<\n");
      exit(-1);
      }
#endif
   }

int main(int argc, char *argv[])
{
   int arg,c, sd, s_n, daemon, bgp_result, k;
   char *ap;
   struct sockaddr_in me;
   struct utsname name;
#if DAG_DIRECT
   int dag_type = TT_EXT;
   int dag_varlen = DAG_DEFAULT_VARLEN;
   int dag_slen = DAG_DEFAULT_SLEN;
   int dag_ports = 1;
#endif
   int hyperthread_processors = 0;
   unsigned meter_mem;
   struct tms tbuf;

   mxpkts = DFMXPKTS;  mxflows = DFMXFLOWS;  mxrules = DFMXRULES;
   mxstreams = DFMXSTREAMS;  mxstr_blocks = DFMXSTRBLOCKS;
   mxdistribs = DFMXDISTRIBS;  mxdistevents = DFMXEVENTS;
   mxpktdatas = DFMXPKTDATS;  mxoamrecs = DFMXOAMRECS;
   mxusers = DFMXUSERS;  mxhlists = DFMXHLISTS;

   if (argc < 2) {  /* Help Screen when called with no args */
      print_help(argv[0]);
      check_struct_sizes();
      exit(-1);
      }

   signal(SIGINT, sigint_handler);
   signal(SIGTERM, sigint_handler);
   signal(SIGUSR1, sigusr1_handler);
   signal(SIGUSR2, sigusr2_handler);

#if CR_DATA
   use_ip_length = 1;  /* Use IP header lengths */
   coral_set_api(CORAL_API_CELL);  /* Allows config file with proto rules */
#elif DAG_DATA || DAG_DIRECT
   use_ip_length = 1;  /* Use IP header lengths */
#endif

   display_enabled = kb_enabled = 1;  /* Enabled by default */
   daemon = 0;
   /* Default (CMU) communities are:  0 = "public", 1 = "proxy",
	 2 = "private", 3 = "regional", 4 = "core"
      We only allow "public" and "private" by default */
   communities[1] = communities[3] = communities[4] = "";
   for (n_interfaces = 0;  n_interfaces != MXINTERFACES; ++n_interfaces)
      pci[n_interfaces] = NULL;
   s_n = 1;  n_interfaces = 0;
   for (c = 0, arg = 1; arg < argc; ++arg) {
      if (argv[arg][0] == '-') {
         ap = argv[arg]+2;
         switch (argv[arg][1]) {
         case 'a':
            if (*ap == '\0') ap = argv[++arg];
            mxpktdatas = atoi(ap);
            break;
	 case 'b':
	    if (*ap == '\0') ap = argv[++arg];
	    mxstr_blocks = atoi(ap);
	    break;
	 case 'd':
	    snmp_dump_packet++;
	    break;
         case 'e':
            if (*ap == '\0') ap = argv[++arg];
            mxdistevents = atoi(ap);
            break;
     	 case 'f':
	    if (*ap == '\0') ap = argv[++arg];
	    mxflows = atoi(ap);
	    break;
#if DAG_DIRECT
         case 'g':  /* -g dag type, default = 0 (atm) */
            if (*ap == '\0') ap = argv[++arg];
            dag_type = atoi(ap);
            if (dag_type < TT_ATM || dag_type > TT_EXT)
               log_msg(LOG_ERR, 1,
                  "Dag type (-g) not 0 (ATM), 1 (Ether) 2 (PoS) or 3 (Ext)\n");
            break;
#endif
	 case 'i':  /* -i name of interface to monitor */
	    if (*ap == '\0') ap = argv[++arg];
            pci[n_interfaces] = (struct interface_info *)calloc(
               sizeof(struct interface_info), 1);
            strncpy(pci[n_interfaces]->name,ap,SZ_IFNAME);
            pci[n_interfaces]->name[SZ_IFNAME] = '\0';
#if DAG_DIRECT
            pci[n_interfaces]->dag_type = dag_type;
            pci[n_interfaces]->dag_slen = dag_slen;
            pci[n_interfaces]->dag_varlen = dag_varlen;
            pci[n_interfaces]->dag_ports = dag_ports;
#endif
            pci[n_interfaces]->nbr = n_interfaces+1;  /* 1-org */
            ++n_interfaces;
	    break;
	 case 'k':
	    kb_enabled = 0;  /* -k to disable keyboard */
	    break;
      	 case 'l':  /* Use 'Total Length' field from IP datagrams */
	    use_ip_length = 1;
	    break;
	 case 'm':
	    if (*ap == '\0') ap = argv[++arg];
            au_snmp_port = atoi(ap);
            break;
 	 case 'n':
 	    if (*ap == '\0') ap = argv[++arg];
            if (ap != NULL) s_n = atoi(ap);
            else s_n = 1;
            if (s_n < 1) s_n = 1;
 	    break;
#if NF_OCX_BGP
	 case 'o':
	    use_owner_asns++;
	    break;
#endif
#ifdef LFAP_METER
 	 case 'p':
 	    if (*ap == '\0') ap = argv[++arg];
		if (ap!=NULL) lfap_info_level = atoi(ap);
		else lfap_info_level = LFAP_INFLVL_FLOW|LFAP_INFLVL_MSG;
            if (lfap_info_level < 0) lfap_info_level = 0;
/*          else if (lfap_info_level > 15) lfap_info_level = 15; */
 	    break;
#endif
	 case 'r':
	    if (*ap == '\0') ap = argv[++arg];  /* -r to set read community */
            strncpy(r_community,ap,SZ_IFNAME);
            r_community[SZ_IFNAME] = '\0';
            communities[0] = r_community;
	    break;
	 case 's':
	    display_enabled = 0;  /* -s to disable screen */
	    break;
         case 't':
            if (*ap == '\0') ap = argv[++arg];
            mxstreams = atoi(ap);
            break;
	 case 'u':
	    if (*ap == '\0') ap = argv[++arg];
	    mxrules = atoi(ap);
	    break;
 	 case 'v':
            if (*ap == '\0') ap = argv[++arg];
            mxdistribs = atoi(ap);
	    break;
	 case 'w':
	    if (*ap == '\0') ap = argv[++arg];  /* -w to set write community */
            strncpy(w_community,ap,SZ_IFNAME);
            w_community[SZ_IFNAME] = '\0';
 	    communities[2] = w_community;
            if (strcmp(w_community, "public") == 0) {
                /* 'public' is the default for the readonly community, set
                    it to somthing else, otherwise snmp's get_community will
                    trip over 'public' first, ane never get write access! */
               strcpy(r_community,"PUBLIC");
               communities[0] = r_community;
               }
	    break;
#if CR_DATA
         case 'C':  /* CoralReef option (other than source) */
            if (*ap == '\0') ap = argv[++arg];
            if (coral_config_command(ap) < 0)
                exit(-1);
            break;
#endif
         case 'D':
	    if (!trace_interval) {
               display_enabled = kb_enabled = 0;
               daemon++;
	       }
            else log_msg(LOG_WARNING, 0,
               "-D options reqires -T0 to run as daemon!");
            break;
     	 case 'H':
	    if (*ap == '\0') ap = argv[++arg];
	    mxhlists = atoi(ap);
	    break;
         case 'M':  /* Specify name for meter.log file */
	    if (*ap == '\0') ap = argv[++arg];
            logname = malloc(strlen(ap)+1);
            strcpy(logname, ap);
            break;
#if CR_DATA || DAG_DIRECT
         case 'N':  /* Stop after N*T s of tracefile.  0 => stop at EOF */
            if (*ap == '\0') ap = argv[++arg];
            crl_n_max = atoi(ap);
            break;
#endif
     	 case 'O':
	    if (*ap == '\0') ap = argv[++arg];
	    mxoamrecs = atoi(ap);
	    break;
         case 'P':
            if (*ap == '\0') ap = argv[++arg];
            mxpkts = atoi(ap);
            break;
#ifdef LFAP_METER
 	 case 'R':
 	    if (*ap == '\0') ap = argv[++arg];
            strncpy(lfap_community,ap,SZ_IFNAME);
            lfap_community[SZ_IFNAME-1] = '\0';
 	    break;
#endif
#if CR_DATA
         case 'S':  /* CoralReef source */
            if (*ap == '\0') ap = argv[++arg];
            pci[n_interfaces] = (struct interface_info *)calloc(
               sizeof(struct interface_info), 1);
            strncpy(pci[n_interfaces]->name,ap,SZ_IFNAME);
            pci[n_interfaces]->name[SZ_IFNAME] = '\0';
            pci[n_interfaces]->nbr = n_interfaces+1;  /* 1-org */
            pci[n_interfaces]->crl_src = coral_new_source(ap);
            if (pci[n_interfaces]->crl_src == NULL) {
               printf("Failed to open coral source %s\n", ap);
               exit(-1);
               }
            ++n_interfaces;
            break;
#endif
#if CR_DATA || DAG_DIRECT
         case 'T':  /* Collection interval (seconds) for trace file.
                       T0 (default) implies 'use live source(s)' */
            if (*ap == '\0') ap = argv[++arg];
	    trace_interval = atoi(ap);
            break;
#endif
#if DAG_DIRECT
         case 'F':
            dag_varlen = 0;
            break;
         case 'I':
            if (*ap == '\0') ap = argv[++arg];
            dag_ports = atoi(ap);
            break;
         case 'L':
            if (*ap == '\0') ap = argv[++arg];
            dag_slen = atoi(ap);
            break;
         case 'V':
            dag_varlen = 1;
            break;
#endif
#if USER_INFO
         case 'U':
            if (*ap == '\0') ap = argv[++arg];
            mxusers = atoi(ap);
            break;
#endif
         case 'X':
            ++hyperthread_processors;
            break;
	 default:
	    log_msg(LOG_ERR, 0, "Invalid option: -%c", argv[arg][1]);
         case '?':
         case '-':
         case 'h':
            print_help(argv[0]);
            exit(-1);
	    }
	 }
      else log_msg(LOG_ERR, 1, "Option not preceded by -\n");
      }

   if (daemon) {  /* User asked to run NeTraMet as daemon */
      switch (k = fork()) {
      case 0:  /* Child (left running on it's own) */
         break;
      case -1:
         fprintf(stderr, "Fork failed!!\n");
         exit(1);
      default:  /* Parent */
         exit(0);
         }
      setsid();  /* Move into a new session */
      }

#if NF_OCX_BGP
    InitSubnet();
    bgp_result = read_bgp_file("bgp.txt");
    if (bgp_result) {
       log_msg(LOG_ERR, 1,
          "read_bgp_file failed, bgp_result=%d\n", bgp_result);
       }
#endif

   /* Set up connections */
   sd = socket(AF_INET, SOCK_DGRAM, 0);
   if (sd < 0)
      log_msg(LOG_ERR, 1, "Couldn't get socket");
   me.sin_family = AF_INET;
   me.sin_addr.s_addr = INADDR_ANY;
   me.sin_port = htons(SNMP_PORT);
   if (bind(sd, (struct sockaddr *)&me, sizeof(me)) != 0)
      log_msg(LOG_ERR, 2,  /* SNMP_PORT defined in snmplib/snmp.h */
         "Couldn't bind port %d: %s", SNMP_PORT,strerror(errno));
   if (n_interfaces == 0) {
      pci[0] = 
         (struct interface_info *)calloc(sizeof(struct interface_info),1);
      pci[0]->nbr = n_interfaces = 1;
      }
   for (c = 0; c != n_interfaces; ++c)
      pci[c]->SampleRate = pci[c]->sample_count = s_n;

#ifdef LFAP_METER
   if (!start_lfap(&name)) return 0;  /* Fail */
#endif

   clear_pkt_stats = 1;  sproctime = 0;
   get_sys_info(hyperthread_processors);
   clocktime_org = sclocktime = times(&tbuf);
   av_idle1000 = proctime_org = sproctime = 0;

#if MULTI
   init_mutexes();

   ntm_semaphore_init("osp_semaphore",0, &osp_semaphore);
   semaphore_decrement(&osp_semaphore);  /* Thread blocked */
   if (gprof_pthread_create(&osp_thread, NULL, run_osp, NULL)) {
      log_msg(LOG_ERR, 1, "Error creating osp thread");
      }

   for (k = 0; k != N_PME_THREADS; ++k) {
      struct pme_data *pd = &pme_data_buf[k];
      memset(pd, 0, sizeof(struct pme_data));
      ntm_semaphore_init("buf_full",k+1, &pd->buf_full);
      ntm_semaphore_init("buf_empty",k+1, &pd->buf_empty);
      semaphore_decrement(&pd->buf_full);  /* Thread blocked */
      semaphore_decrement(&pd->buf_empty);
      pd->hash_group = k;
      if (gprof_pthread_create(&pd->pme_thread, NULL, run_pme, pd)) {
         log_msg(LOG_ERR, 1, "Error creating pme %d thread", k);
         }
      }

   /* Allow init_monitor() to open default ruleset */
   ntm_semaphore_init("running_rulesets",0, &r_rts_sem);
   semaphore_decrement(&r_rts_sem);
   for (k = 0; k != N_PME_THREADS; ++k)
      ntm_semaphore_up("init", &r_rts_sem);
#endif

   meter_mem = init_monitor(n_interfaces);
   init_snmp();
   uname(&name);

   log_msg(LOG_WARNING, 0, "Starting %s", version_descr);
   phys_mem = phys_mem/(1024*1024);  meter_mem = meter_mem/(1024*1024);
   log_msg(LOG_WARNING, 0,
      "Running on %s (%u processors, %u MiB memory), using %u MiB", 
      name.nodename, n_procs, phys_mem, meter_mem);
   if ((k = meter_mem*100/phys_mem) > 50) log_msg(LOG_WARNING, 0,
      "Meter is using %u%% of pysical memory, beware of paging load", k);

   ap = interface_name;  k = sizeof(interface_name);
   for (c = 0; c != n_interfaces; ++c) {
      if (!init_interface(pci[c])) exit(3);
      if (strlen(pci[c]->name)+3 < k) {
         if (c != 0) ap = strmov(ap,", ");
         ap = strmov(ap,pci[c]->name);
         k -= strlen(pci[c]->name)+2;
         }
      }
   *ap = '\0';

#if NETFLOW
   log_msg(LOG_WARNING, 0, "Metering port(s) %s", interface_name);
#else
   log_msg(LOG_WARNING, 0, "Metering interface(s) %s", interface_name);
#endif

   setuid(getuid());
   receive(sd,n_interfaces);
   }

void zero_pkt_stats()
{
   struct tms tbuf;
#if LOG_THREAD_PROCTIMES
   time_t t;  char *ts, ftb[32];
   int j;
   Bit32 pt, pc_proc;

   time(&t);  ts = ctime(&t);
   strncpy(&ftb[0],  &ts[11], 9);  /* 17:31:42_ */
   strncpy(&ftb[9],  &ts[0],  4);  /* Thu_ */
   strncpy(&ftb[13], &ts[8],  3);  /* 23_  */
   strncpy(&ftb[16], &ts[4],  4);  /* Sep_ */
   strncpy(&ftb[20], &ts[20], 4);  /* 1997 */
   ftb[24] = '\0';

   if (threads_log == NULL && sproctime != 0) {
      if ((threads_log = fopen("threads.log", "a")) == NULL)
         log_msg(LOG_ERR, 1, "Couldn't open threads.log <<<<<");
      fprintf(threads_log, "%s #Threads: 1=%s", ftb, thread_name[0]);
      for (j = 1; j != t_log_n; ++j)
         fprintf(threads_log, " %d=%s", j+1, thread_name[j]);
      fprintf(threads_log, "\n");
      }

   if(threads_log != NULL) {
      fprintf(threads_log, "%s ", ftb);
      pt = sclocktime-clocktime_org;
      if (pt == 0) pt = 1;  /* Don't divide by 0 */
      for (j = 0; j != t_log_n; ++j) {
         fprintf(threads_log, " %d", (thread_proctime[j]-thread_proctime_org[j])*1000/pt);
         thread_proctime_org[j] = thread_proctime[j];
         }
      fprintf(threads_log, "\n");
      fflush(threads_log);
      }
#endif   

   clocktime_org = sclocktime;  proctime_org = sproctime;
   min_idle1000 = 1000L;
   noflowpackets = 0;
   pkt_counts(1);  /* Reset origins */
   stats_time = spackets = 0L;
   max_pkt_rate = 0;
   clear_pkt_stats = 0;

   n_lost_streams = n_lost_flows = 0;
   mx_active_streams = mx_active_flows = 0;
   }

int chart_interface = 0;  /* Use SampleRate from first interface */
unsigned int bkgi;  /* Seconds before next run of backgound process */
#define BKG_INTERVAL  30  /* Seconds */
int gci;

#if CR_DATA || DAG_DIRECT
unsigned int mtrlogi = 30;  /* Seconds before next run of meter log process */
/* #define LOG_INTERVAL  60  /* Seconds (1 min) */
#define LOG_INTERVAL  300  /* Seconds (5 min) */
#endif

void one_second_process(void)
{
   int x;
   clock_t rt,pt, clocktime, proctime;
   struct tms tbuf;
   unsigned long idle1000, pd, cd;
   unsigned char p;
   struct mgr_rec *mip;
   struct rdr_rec *cip;
   char msg[100];
#if CR_DATA || DAG_DIRECT
   int i, ifd, flags;
   char *mp, tsbuf[30], ibuf[40];
#endif
   Bit32 active;
   Bit32 osp_proct;

#if UX_TESTING
   printf("+");
#endif
   if (request_stop) stop_meter(0);

   ++stats_time;
   clocktime = times(&tbuf);  /* Times for self + timed-out children */
#if !MULTI
   proctime = tbuf.tms_utime+tbuf.tms_stime;
#else
   osp_proct = tbuf.tms_utime+tbuf.tms_stime;
   total_thread_proct += (osp_proct - last_osp_proct);
# if LOG_THREAD_PROCTIMES
   thread_proctime[osp_tn] += (osp_proct - last_osp_proct);
# endif   
   last_osp_proct = osp_proct;
   proctime = total_thread_proct;
#endif

   if (clocktime > sclocktime) {
      pd = proctime-sproctime;  cd = n_procs*(clocktime-sclocktime);
      if (pd <= cd) {
         idle1000 = 1000L - pd*1000L/cd;
         if (idle1000 < min_idle1000) min_idle1000 = idle1000;
         pd = proctime-proctime_org;  cd = n_procs*(clocktime-clocktime_org);
         if (pd <= cd)
            av_idle1000 = 1000L - pd*1000L/cd;
         }
      /* printf("last_t_s=%u, idle=%u, min_idle=%u, av_idle=%u\n",
        last_t_sec, idle1000, min_idle1000, av_idle1000);  /**/
      }
   sclocktime = clocktime;  sproctime = proctime;
   if (clear_pkt_stats) zero_pkt_stats();

   pkt_counts(0);

#if 0
   if (stats_time != 0) { //19
      printf("--- osp(%d):  stats_time=%lu, npackets=%lu, rate=%.2f kp/s \n", 
         last_t_sec,  stats_time, npackets,
         (double)npackets/(1000.0*stats_time) );
      fflush(stdout);
      }
#endif
   pd = (npackets-spackets)*pci[chart_interface]->SampleRate;
   if (pd > max_pkt_rate) max_pkt_rate = pd;
   spackets = npackets;

   active = mxstreams - n_free_streams;
   if (active > mx_active_streams) mx_active_streams = active;
   active = active_flows();
   if (active > mx_active_flows) mx_active_flows = active;

   check_events(uptime_s());  /* uptime in seconds */

   if (--gci == 0) {
#if UX_TESTING
     printf("g");
#endif
      garbage_collect(1);  /* Routine incremental collection */
      gci = gc_interval;
      }

#if CR_DATA || DAG_DIRECT
   if (--mtrlogi == 0 && live_source) {  /* Write regular meter log entry */
      int cpb, mxcpb;
      for (mp = msg, mxcpb = i = 0; i != n_interfaces; ++i) {
         cpb = pci[i]->nblocks == 0 ? 0 :
            pci[i]->nrecords/pci[i]->nblocks;
         fmt_ts(tsbuf, pci[i]->card_ts);
         sprintf(ibuf, "%d:  %s, %lu badts, %lu cpb, %lu_%lu^%lu, ",
            i+1, tsbuf, pci[i]->badtspkts, cpb, 
            pci[i]->nblocks, pci[i]->mx_empty, pci[i]->mx_full);
# if DAG_DIRECT
         if (pci[i]->nblocks == 0) {  /* Dead interface */
            log_msg(LOG_ERR, 0, "Try to restart i/f %d", i+1);
            if (pci[i]->restarts) log_msg(LOG_ERR, 111,
               "No data after previous restart, shutting down");
            else if (open_dag_direct(pci[i])) {
               log_msg(LOG_ERR, 0, "Restarted OK");
               ++pci[i]->restarts;
               }
            else log_msg(LOG_ERR, 112, "Restart failed, shutting down");
	    }
# endif
         pci[i]->nrecords = pci[i]->nblocks =
            pci[i]->mx_empty = pci[i]->mx_full = 0;  
         if (cpb > mxcpb) mxcpb = cpb;
         mp = strmov(mp, ibuf);
         }
      mp[-2] = '\0';
      log_msg(LOG_WARNING, 0, "%s. %u delta_OKs", msg, n_delta_OKs);
      if (n_delta_OKs < 100 && last_n_delta_OKs > 500) {
         log_msg(LOG_ERR, 113,
            "Not enough matched packet pairs, shutting down");
         }
      else {
         last_n_delta_OKs = n_delta_OKs;  n_delta_OKs = 0;
         }
/*      show_stream_hash(0);  /*PPP*/
      mtrlogi = LOG_INTERVAL;

      if (mxcpb < TARGET_CPB) {  /* Should we adjust the select time? */
         if (select_us < MX_SELECT_US) {
            select_us *= 2;
# if DAG_DIRECT
            select_time = US_DAGT(select_us);
# elif CR_DATA
            select_tv.tv_usec = select_us;
# endif
            log_msg(LOG_WARNING, 0, "--* select_us = %d", select_us);
            }
         }
      else if (mxcpb >= 2*TARGET_CPB) {
         if (select_us > MN_SELECT_US) {
            select_us /= 2;
#if DAG_DIRECT
            select_time = US_DAGT(select_us);
#elif CR_DATA
            select_tv.tv_usec = select_us;
#endif
            log_msg(LOG_WARNING, 0, "--/ select_us = %d", select_us);
            }
         }
      }
#endif

   if (--bkgi == 0) {  /* Check % of flows active */
#if UX_TESTING
      printf("b\n");
#endif
      p = (unsigned long)active_flows()*100/(mxflows);

      if (p > FloodMark && FloodMode != TV_TRUE &&
            FloodMark > 0 && FloodMark < 100) {
         FloodMode = TV_TRUE;  /* Stop creating new flow records */
         log_msg(LOG_WARNING, 0, "Meter in Flood mode");
	 }
      else for (x = 0; x != sizeof(mi)/sizeof(struct mgr_rec); ++x) {
         mip = &mi[x];
         if (mip->mi_Status != RS_ACTIVE
               || mip->mi_RunningStandby == TV_TRUE)
            continue;
	 else if (p > mip->mi_HighWaterMark &&
	       mip->mi_HighWaterMark > 0 && 
               mip->mi_HighWaterMark < 100 &&
	       mip->mi_StandbyRuleSet != mip->mi_CurrentRuleSet &&
               (mip->mi_StandbyRuleSet != 0 &&
                  ri[mip->mi_StandbyRuleSet-1].ri_Status == RS_ACTIVE)
               ) {
            open_rule_set(mip->mi_CurrentRuleSet, 0);  /* Close */
	    open_rule_set(mip->mi_StandbyRuleSet, 1);  /* Open */
            mip->mi_RunningStandby = TV_TRUE;
            log_msg(LOG_WARNING, 0, "Manager %d running Standby", x+1);
	    }
	 else if (p > HighWaterMark*2/3)
	    more_garbage();
         }
      pd = uptime_cs();
      for (x = 0; x != sizeof(ci)/sizeof(struct rdr_rec); ++x) {
         cip = &ci[x];
         if (cip->ci_Status != RS_ACTIVE || cip->ci_Timeout == 0)
            continue;
         if ((pd - cip->ci_LastTime)/100 > cip->ci_Timeout)
            cip->ci_Status = RS_UNUSED;  /* Timed out */
         }
      bkgi = BKG_INTERVAL;
      }
#if UX_TESTING
   fflush(stdout);
#endif
   }

handle_keyboard()
{
   char kb_buf[25];
   int ch;
   if (fgets(kb_buf, sizeof kb_buf, stdin) == NULL)
      return;  /* Error or EOF */
   if ((ch = kb_buf[0]) == 27 && kb_buf[1] == 27)  /* ESC ESC */
      stop_meter(0);
   switch (tolower(ch)) {
   case 'v':
      printf("\n%s\n",version_descr);
      break;
#if CR_DATA
   case '*':
      if (coral_start_all() < 0)  /* Also synchronises all sources */
         log_msg(LOG_ERR, 0, "coral_start failed");
      break;
#endif
   default:
      handle_kb(ch);
      break;
      }
   }

void snmp_read_sp(int sd);  /* Handles meter's SNMP requests */

#if CR_DATA || DAG_DIRECT

int non_live_source(int sd)  /* New (MULTI) version, runs as OB */
{
   fd_set fdset;
   struct timeval t;
   int j, eof, width, count;
   int last_state = -1;
#if DEBUG_TF_READ
   char tsbuf[30];
#endif

   width = STDIN_FILENO;  /* Determine size of fd_set */
   if (sd > width) width = sd;
   ++width;
   FD_ZERO(&fdset);
   t.tv_sec = 0;  

   for (;;) {
#if DEBUG_TF_READ
      if (trace_state != last_state) {
         fmt_ts(tsbuf, last_ts);
         printf("Main loop: state=%d, crl_n_count=%d, last_t_sec=%d"
            ", last_ts=%s\n",
            trace_state, crl_n_count, last_t_sec, tsbuf);
         }
#endif

      switch (trace_state) {  /* Reading from trace file */
      case CT_WAIT_RULES:
         /* met_vars will change to CT_RULES_READY */
         break;
      case CT_RULES_READY:
         trace_state = CT_WAIT_FLOWS_READ;
         break;
      case CT_PROCESS_FLOWS:
            ntm_semaphore_down("nls", &merge_sync_empty);

         for (eof = j = 0; j != n_interfaces; ++j) {
#if HAVE_LIBZ
            if (pci[j]->gfp == NULL) {
#else
            if (pci[j]->t_fd == -1) {
#endif
               trace_state = CT_TRACE_EOF;  /* Dummy up EOF */
               break;
               }
            }
         if (trace_state == CT_TRACE_EOF) break;

         if (trace_count <= 0) {  /* Reached -T interval */
/*            show_memory();  /* Show memory every -T interval */
/*            show_stream_hash();  */
            trace_count += trace_interval;
            if (crl_n_max) {  /* Stop after -N meter readings */
               if (++crl_n_count == crl_n_max) {
                 trace_state = CT_TRACE_EOF;  /* Dummy up EOF */
                 }
               else
                  trace_state = CT_WAIT_FLOWS_READ;
               }
            else trace_state = CT_WAIT_FLOWS_READ;
            }

            ntm_semaphore_up("nls", &merge_sync_full);
         break;
      case CT_WAIT_FLOWS_READ:
         /* NeMaC will change to CT_PROCESS_FLOWS */
         break;
      case CT_TRACE_EOF:
         /* NeMaC will change to CT_SHUTDOWN */
         break;
      case CT_SHUTDOWN:
         stop_meter(0);
         break;
         }
      last_state = trace_state;

      t.tv_usec = 10000;  /* 0.01 s  Keep snmp and keyboard alive */
      FD_SET(sd, &fdset);
      if (kb_enabled) FD_SET(STDIN_FILENO, &fdset);  /* stdin */
      count = select(width, &fdset, 0, 0, &t);
      if (count > 0) {
         if (FD_ISSET(sd, &fdset))
	    snmp_read_sp(sd);
         if (kb_enabled && FD_ISSET(STDIN_FILENO, &fdset))  /* stdin */
            handle_keyboard();
         }
      else switch (count) {
      case 0:
	 break;
      case -1:
	 if (errno == EINTR) continue;
	 else log_msg(LOG_ERR, 0,
            "select returned %d, %s", count, strerror(errno));
	 return 0;
      default:
	 log_msg(LOG_ERR, 0, "select returned %d", count);
	 return 0;
	 }

      }
   }

# if 0  /* Old version */
int non_live_source(int sd)
{
#if CR_DATA
   int last_trace_count;
#endif

   width = STDIN_FILENO;  /* Determine size of fd_set */
   if (sd > width) width = sd;
   ++width;
   FD_ZERO(&fdset);
   t.tv_sec = 0;  

   for (;;) {
#if DEBUG_TF_READ
      if (trace_state != last_state) {
         fmt_ts(tsbuf, last_ts);
         printf("Main loop: state=%d, crl_n_count=%d, last_t_sec=%d"
            ", last_ts=%s\n",
            trace_state, crl_n_count, last_t_sec, tsbuf);
         last_state = trace_state;
         }
#endif
      t.tv_usec = 250000;  /* 0.25 s */
      switch (trace_state) {  /* Reading from trace file */
      case CT_WAIT_RULES:
         /* met_vars will change to CT_RULES_READY */
         break;
      case CT_RULES_READY:
#if CR_DATA
         interface_read_trace(pci[0]);
            /* Get timestamp for beginning of first interval */
         last_t_sec = last_ts.tv_sec;
         meter_start_time.tv_sec = last_ts.tv_sec-1;
            /* Don't let uptime start at -1 */
#endif
         trace_state = CT_WAIT_FLOWS_READ;
         break;
      case CT_PROCESS_FLOWS:
#if CR_DATA
         t.tv_usec = 20000;  /* 20 ms */
         last_trace_count = trace_count;
         do {
            if (!interface_read_trace(pci[0])) {
               trace_state = CT_TRACE_EOF;  /* No more blocks */
               break;
               }
            } while (trace_count != 0);
         if (trace_state == CT_TRACE_EOF)
            break;  /* Let NeMaC do EOF processing (i.e. last reading) */

#elif DAG_DIRECT && !DAG_NEW_API
         dag_merge();  /* Return at end of block */
         if (pci[waiting_live_source]->r_bytes == 0) {  /* Block now empty */
            if (read_live_block(waiting_live_source) == 0) {
               trace_state = CT_TRACE_EOF;  /* No more blocks */
               break;
               }
            }
#endif
         if (trace_count == 0) {  /* Reached -T interval */
/*            show_memory();  /* Show memory every -T interval */
/*            show_stream_hash();  */
            trace_count = trace_interval;
            if (crl_n_max) {  /* Stop after -N meter readings */
               if (++crl_n_count == crl_n_max) {
                 trace_state = CT_TRACE_EOF;  /* Dummy up EOF */
                 }
               else
                  trace_state = CT_WAIT_FLOWS_READ;
               }
            else trace_state = CT_WAIT_FLOWS_READ;
            }
         break;
      case CT_WAIT_FLOWS_READ:
         /* NeMaC will change to CT_PROCESS_FLOWS */
         break;
      case CT_TRACE_EOF:
         /* NeMaC will change to CT_SHUTDOWN */
         break;
      case CT_SHUTDOWN:
         stop_meter(0);
         break;
         }

      FD_SET(sd, &fdset);
      if (kb_enabled) FD_SET(STDIN_FILENO, &fdset);  /* stdin */
      count = select(width, &fdset, 0, 0, &t);
      if (count > 0) {
         if (FD_ISSET(sd, &fdset))
	    snmp_read_sp(sd);
         if (kb_enabled && FD_ISSET(STDIN_FILENO, &fdset))  /* stdin */
            handle_keyboard();
         }
      else switch (count) {
      case 0:
	 break;
      case -1:
	 if (errno == EINTR) continue;
	 else log_msg(LOG_ERR, 0,
            "select returned %d, %s", count, strerror(errno));
	 return 0;
      default:
	 log_msg(LOG_ERR, 0, "select returned %d", count);
	 return 0;
	 }
      }
   return 1;
   }
# endif  /* Old version */
#endif  /* CR_DATA || DAG_DIRECT */

#if MULTI
# if DAG_DIRECT
void *run_dag_merge(void *arg)
{
   struct timeval timeout;

   init_hist(&dm_hist, "dag_mg");
   dag_merge();  /* Just keep running */
#  if 0  /* select() doesn't work properly in linux/pthreads! */
   for (;;) {
         /* Block until semaphore comes up, -> pme buffer empty */
      dag_merge();  /* Return when ready for next block */

      while (read_live_block(waiting_live_source)) {
         timeout.tv_sec = 0;  timeout.tv_usec = select_us;
         select(FD_SETSIZE, 0, 0, 0, &timeout);
         }
      }
   return NULL;  /* Keep compiler happy */
#  endif
   }
# endif
#endif


int receive(int sd,int n_interfaces)
{
   fd_set fdset;
   int j, width, count;
   struct timeval t;
#if NETFLOW || CR_DATA || LFAP_METER
   time_t now,prev;
#endif
#if DAG_DIRECT
   int i_count = 0;
#endif
# if MULTI
   pthread_t pkt_read_thread;
   Bit32 o_b_proct,
      last_o_b_proct = 0;
   struct tms tbuf;
# endif

#define OB_TIMING_TEST  0
#if OB_TIMING_TEST
   struct timeval t_ob;
   double last_d_ob, d_ob;
#endif

#if WATCH_RUN_LIGHT
   struct timeval rlt;
   int last_rlt_sec = 0;
   int n_rl_warns = 0;
#define RUN_LIGHT_MAX_TIME  60  /* Seconds */
   /* Two Dag cards can start a minute apart, and take a few
      seconds for the busy one to catch up, that's why we
      wait so long before saying 'no packets!'  */
#define RUN_LIGHT_MAX_WARNS 3  /* exit() after 3 MAX_TIMEs */
#endif

#if LOG_THREAD_PROCTIMES
   int this_tn = assign_thr_log_nbr();
   sprintf(thread_name[this_tn], "outer_blk");
   thread_proctime_org[this_tn] = thread_proctime[this_tn] = 0;
#endif   
/*   printf("running receive():\n");  fflush(stdout);  /* */

   gci = gc_interval;  /* Garbage collect interval */
   bkgi = BKG_INTERVAL;  /* Background interval */
   gc_f = (mxflows)/100;  /* Flow indexes to check */
   if (gc_f < 4) gc_f = 4;

#if DAG_DIRECT
   select_time = US_DAGT(select_us);
#endif

#if CR_DATA
   init_live_sources();  /* Read some cells from each interface */
   coral_merge();  /* Return when ready for next block */
#elif DAG_DIRECT

# if MULTI  /* Run dag_merge process in its own thread */
   if (!live_source) {
      ntm_semaphore_init("merge_sync_full",0, &merge_sync_full);
      ntm_semaphore_init("merge_sync_empty",0, &merge_sync_empty);
      semaphore_decrement(&merge_sync_full);  /* Thread blocked */
      semaphore_decrement(&merge_sync_empty);
      }
   init_live_sources();  /* Read first block, set meter_start_time */
   if (gprof_pthread_create(&pkt_read_thread, NULL, run_dag_merge, NULL)) {
      log_msg(LOG_ERR, 1, "Error creating read thread");
      }
# else
   dag_merge();  /* Return when ready for next block */
# endif
#endif

#if MULTI && !DAG_DIRECT   /* Start 'merge' thread */
   ntm_semaphore_init("if_buff_full",0, &if_buf_full);
   ntm_semaphore_init("if_buf_empty",0, &if_buf_empty);
   semaphore_decrement(&if_buf_full);  /* Thread blocked */
   semaphore_decrement(&if_buf_empty);
   if (gprof_pthread_create(&pkt_read_thread, NULL, run_if_read, NULL) ) {
      log_msg(LOG_ERR, 1, "Error creating read thread");
      }
#endif

#if CR_DATA || DAG_DIRECT
   if (!live_source)
      non_live_source(sd);
#endif

   width = STDIN_FILENO;  /* Determine size of fd_set */
   if (sd > width) width = sd;
#if !CYGWIN32
   for (j = 0; j != n_interfaces; ++j)
      if (pci[j]->fd > width) width = pci[j]->fd;
#endif
   ++width;
#if NETFLOW || LFAP_METER
   time(&prev);
#endif

#if OB_TIMING_TEST
   gettimeofday(&t_ob, (struct timezone *)0);
   last_d_ob = t_ob.tv_sec + t_ob.tv_usec/1000000;
#endif
   for (;;) {
#if WATCH_RUN_LIGHT
      if (got_USR2) {
         printf(".");  fflush(stdout);
         }
      gettimeofday(&rlt, (struct timezone *)0);
      if (rlt.tv_sec > last_rlt_sec) {
         if (rlt.tv_sec - last_t_sec > RUN_LIGHT_MAX_TIME) {
            log_msg(LOG_WARNING, 0,
               "Run light out for %ds, warning %d",
               rlt.tv_sec-last_t_sec, n_rl_warns);
            if (++n_rl_warns == RUN_LIGHT_MAX_WARNS)
               log_msg(LOG_ERR, 111, "Stopped reading packets <<<<");
# if DAG_HW_RESET
            for (j = 0; j != n_interfaces; ++j) {
               if (pci[j]->nblocks == pci[j]->last_nblocks &&
                     pci[j]->running) {  /* No blocks read for 2s */
                  log_msg(LOG_WARNING, 0,
                     "No blocks read for 2s, i/f %d", j+1);
                  dag_reset(j);  /* Try to reset the Dag card */
                  }
               pci[j]->last_nblocks = pci[j]->nblocks;
               }
# endif
            }
         else n_rl_warns = 0;
         last_rlt_sec = rlt.tv_sec;
         }
#endif  /* WATCH_RUN_LIGHT */

      FD_ZERO(&fdset);
      t.tv_sec = 0;  

#if CR_DATA
      if (read_live_block(waiting_live_source))
         coral_merge();  /* Return when ready for next block */
      else
         if (request_stop) stop_meter(0);
      t.tv_usec = select_us;

#elif DAG_DIRECT  /* !CR_DATA */
      if (request_stop) stop_meter(0);
#if MULTI
      t.tv_usec = 5000;  /* 5 ms */
#else
      if (read_live_block(waiting_live_source)) {
         dag_merge();  /* Return when ready for next block */
         }
      t.tv_usec = select_us;
#endif

#elif CYGWIN32  /* !DAG_DIRECT */
      t.tv_usec =  15000;  /* 15 ms */
      for (j = 0; j != n_interfaces; ++j)  /* Read interfaces every cycle */
         interface_read(pci[j]);  /* Can block for several seconds !!!!!! */

#elif LFAP_METER /* LFAP_METER */
      t.tv_usec = 250000;  /* 0.25 s */
      for (j = 0; j != n_interfaces; ++j) {
         tick(pci[j]);
         if (pci[j]->fd != -1) {
            FD_SET(pci[j]->fd,&fdset);
            if (pci[j]->fd >= width) width = pci[j]->fd + 1;
            }
         }
      FD_SET(lfap_server_socket,&fdset);
      if (lfap_server_socket >= width) width = lfap_server_socket + 1;

#else  /* Unix */
      t.tv_usec = 250000;  /* 0.25 s */
      for (j = 0; j != n_interfaces; ++j) FD_SET(pci[j]->fd,&fdset);
#endif

      FD_SET(sd, &fdset);
      if (kb_enabled) FD_SET(STDIN_FILENO, &fdset);  /* stdin */

      count = select(width, &fdset, 0, 0, &t);
      if (count > 0) {
/*  printf("O/B .. t=%d.%03d, count=%d\n",  t.tv_sec,t.tv_usec, count);  /* $$$ */
#if OB_TIMING_TEST
   gettimeofday(&t_ob, (struct timezone *)0);
   d_ob = t_ob.tv_sec + t_ob.tv_usec/1000000;
   printf("\nOB: delta=%.6f ", d_ob-last_d_ob);  fflush(stdout);
   last_d_ob = d_ob;
#endif

#if LFAP_METER
         if (FD_ISSET(lfap_server_socket, &fdset)) {
            printf("client wants to connect.\n");
            /* Apparently someone knocking on my door ..
               find a free interface */
            for (j = 0; j < n_interfaces; ++j) {
               printf("trying i/f %i.\n",j);
               if (pci[j]->fd == -1) {
                  printf("i/f %i is free.\n",j);  /* Found a free one */
                  pci[j]->fd = accept(lfap_server_socket,NULL,NULL);
                  pci[j]->fasinfo.conn = pci[j]->fd;
                  j = n_interfaces + 1;
                  }
               }
            if (j == n_interfaces) {
               int sock;
               LFAPHeader *lfap;
               LFAP_IE_Header* ie;

               printf("No interfaces left !\n");
               /* Dirty hack to refuse connections, need this in LFAPv5 ? */
               sock = accept(lfap_server_socket, NULL, NULL);
               /* Assume it is the right one ... */
               lfap = newLFAPMessage(LFAP_OPCODE_AR, LFAP_STATUS_SUCCESS, 1);
               /* Fill in command code IE */
               ie = newIE(LFAP_IE_COMMAND_CODE,4);
               putIEDword(ie, (Bit32)LFAP_CMD_CONNECTION_ACCEPTED, 0);
               addIE(&lfap, ie);
               free(ie);
               sendLFAPMessage(lfap, sock);
               free(lfap);
               close(sock);
               }
	    }
#endif

#if !CYGWIN32 && !CR_DATA
         for (j = 0; j != n_interfaces; ++j) {
# if LFAP_METER
            if(pci[j]->fd != -1) {
               if (FD_ISSET(pci[j]->fd, &fdset)) interface_read(pci[j]);
               }
# elif DAG_DIRECT
            if (FD_ISSET(pci[j]->fd, &fdset)) {
	       printf(">>> select returned from Dag i/f %d\n", j);
	       fflush(stdout);
               }

# else  /* !LFAP_METER */
            if (FD_ISSET(pci[j]->fd, &fdset)) {
#  if MULTI
               if_ix = j;
               ntm_semaphore_up("rec", &if_buf_full);
                  /* Let run_if_read have if_ix */
	       ntm_semaphore_down("rec", &if_buf_empty);
                  /* Block until if_ix free again */
#  else
               interface_read(pci[j]);
#  endif
               }
# endif
	    }
#endif
         if (FD_ISSET(sd, &fdset))
	    snmp_read_sp(sd);
         if (kb_enabled && FD_ISSET(STDIN_FILENO, &fdset))  /* stdin */
            handle_keyboard();
         }
      else switch (count) {
      case 0:
	 break;
      case -1:
	 if (errno == EINTR) continue;
	 else log_msg(LOG_ERR, 0, "select returned %d (not EINTR)", count);
	 return 0;
      default:
	 log_msg(LOG_ERR, 0, "select returned %d", count);
	 return 0;
	 }

#if NETFLOW || LFAP_METER
      time(&now);
      if (now != prev) {
         one_second_process();
         prev = now;
         }
#endif

#if MULTI
      times(&tbuf);
      o_b_proct = tbuf.tms_utime+tbuf.tms_stime;
      total_thread_proct += (o_b_proct - last_o_b_proct);
#if LOG_THREAD_PROCTIMES
      thread_proctime[this_tn] += (o_b_proct - last_o_b_proct);
#endif   
      last_o_b_proct = o_b_proct;
#endif
      }
   return 1;
   }

Bit32 snmp_peer_addr;  /* IPv4 address of SNMP client */

extern n_dds_read;  /* In met_vars.c */
void dyn_dst_init(void);  /* Clear list of dynamic distribs to be reset */

void snmp_read_sp(int sd)
{
   struct sockaddr_in from;
   int length, out_length, fromlength, count;
   Bit8 packet[1500], outpacket[1520];  /* Leave some space just in case */
   char snmp_peer_name[50];  /* Name of host which sent the snmp request */

   fromlength = sizeof from;
   length = recvfrom(sd, packet, 1500, 0, 
      (struct sockaddr *)&from, &fromlength);
   snmp_peer_addr = from.sin_addr.s_addr;
   strcpy(snmp_peer_name,inet_ntoa(from.sin_addr));
   if (length == -1) {
      log_msg(LOG_ERR, 0,
         "snmp_read_sp() recvfrom failed: %s", strerror(errno));
      return;
      }
   if (snmp_dump_packet) {
      printf("received %d bytes from %s:\n", length, snmp_peer_name);
      for (count = 0; count < length; count++) {
         printf("%02X ", packet[count]);
	 if ((count % 16) == 15) printf("\n");
	 else if ((count % 4) == 3) printf(" ");
	 }
      printf("\n\n");
      }

   out_length = 1500;
   if (snmp_agent_parse(packet, length, outpacket, &out_length, 
         from.sin_addr.s_addr)) {
      if (n_dds_read != 0) {
         /* snmp handler leaves dyn_dst mutexes locked if a
            package doesn't fit into the SNMP PDU  */
         dyn_dst_init();  /* We make sure they're not left locked! */
         }
      if (snmp_dump_packet) {
	 printf("sent %d bytes to %s:\n", out_length, snmp_peer_name);
	 for (count = 0; count < out_length; count++) {
            printf("%02X ", outpacket[count]);
	    if ((count % 16) == 15) printf("\n");
	    }
	    printf("\n\n");
	 }
      if (sendto(sd, (char *)outpacket, out_length, 0, 
            (struct sockaddr *)&from, sizeof(from)) < 0) {
         log_msg(LOG_ERR, 0,
            "snmp_read_sp() sendto failed: %s", strerror(errno));
	 }
      }
   }
